
local function lightCallback(fov, x, y)
    return game.lightCallback(fov,x,y)
end

local function trueLightCallback(fov,x,y)
    return true
end

local function checkTile(x, y, r, v)
    map = game._dungeon
    if map[x] and map[x][y] then
        tile = map[x][y]
        if game.debug then
             if not tile.dontColor then
                  tile.color = game.rot.Color._cached.red--color("red")
             end
        end
        if tile.creatures[creature.target.name] then
            creature.hasTarget = tile.creatures[creature.target.name]
            creature.hasTarget._x = game.newPos.x
            creature.hasTarget._y = game.newPos.y
            tile.isPassable = true
        end
    end
end

local function checkTileInRange(x, y, r, v)
    map = game._dungeon
    if map[x] and map[x][y] then
        tile = map[x][y]
        if game.debug then
             if not tile.dontColor then
                  tile.color = game.rot.Color._cached.green--color("red")
             end
        end
        --game:log(game.lume.serialize(creature.target.name).."hmph")
        --game:log(game.lume.serialize(creature.target.tile==tile))
        if tile.creatures[creature.target.name] then
            creature.targetInRange = true game:log("yes in tile")
        end
    end
end

local function tileIsOpen(x, y, r, v)
    map = game._dungeon
    if map[x] and map[x][y] then
        tile = map[x][y]
        if tile.isPassable then
             if not tile.dontColor then
                  tile.color = game.rot.Color._cached.cyan--color("red")
             end
        --    game:log("got a tile to move to")
            table.insert(creature.movementPlaces,tile)
        end
    end
end

--local Creatures = require((...):gsub(('.[^./\\]*'):rep(1) .. '$', ''))

local BaseCreature = Creatures.class:extend("BaseCreature")

function BaseCreature:__init__(kwargs)
    kwargs = kwargs or {}
    
    self.name = kwargs.name or ""
    
    self.health = kwargs.health or 10
    self.strength = kwargs.strength or 10
    self.speed = kwargs.speed or 2
    self.sight = kwargs.sight or 10
    self.attackRange = kwargs.attackRange or 2
    self.hunger = kwargs.hunger or 0
    self.energy = kwargs.energy or 10
    self.movementSpeed = 1
    
    self.hasDied = false
    self._isHunting = false
    self._draw = true
    self._update = true
    self.previousState = {}
    self.repeatTurn = false
    
    self.behaviors = {
    
    }
    
    self.frame_name = "normal"
    self.frame_count = 1
    self.flip = 1
    
    self.foodTarget = {name="player"}
    
    self._x = 0
    self._y = 0
    
    self:init()
end

function BaseCreature:getFrames(n)
    n = n or 4
    frames = {
        normal = {},
        attack = {}
    }
    for i = 1,n do
         --images should be facing right
        table.insert(frames.normal,self.name.."/"..i..".png")
        --attcks are simple for now
        table.insert(frames.attack,self.name.."/".."attack"..i..".png")
    end
    return frames
end

function BaseCreature:init()
    self.frames = self:getFrames(4)
    self.state = "awake"
end

 function BaseCreature:roam()
    self.movementPlaces = {}
    if game.debug then
        log(self.name.." is roaming around.")
    end
    self.movementPlace = nil
    local fov = game.rot.FOV.Precise:new(lightCallback)
    creature = self
    fov:compute(self._x,self._y,self.movementSpeed,tileIsOpen)
    if #self.movementPlaces > 0 then
    
        game:log("movement...")
        self.movementPlace = lume.randomchoice(self.movementPlaces)
        self.path = self:findPathTo(self.movementPlace)
        self._currentPosInPath = 2
        if not self.path then
            game:log("no path")
            return
        end
        self.previousState.move = self.state
        self.state = "moving"
        return true
     end
 end
 
function BaseCreature:isAlive()
    
    if self.health > 0 then
        return true
    elseif self.energy > -9 then
        return true
    end
    return false
end

function BaseCreature:isTired()
   -- if self:isSleeping() then return false end
    if self.energy <= 0 then
        return true
    end
    --if self.state == "sleeping" then
    --    return true
    --end

    --if self.hunger > 10 then
    --    return true
    --end
    --return false
    return false
end

function BaseCreature:isHungry()

  --  if self._isHunting then return false end
    if self.hunger > 10 then
        return true
    end
    return false
end

function BaseCreature:isDead()

    return not self:isAlive()
end

function BaseCreature:isSleeping()

    if self.state == "sleeping" then
        return true
    end
    return false
end

function BaseCreature:isFollowing()

    if self.state == "following" then
        return true
    end
    return false
end

function BaseCreature:isHunting()

    if self.state == "hunting" then
        return true
    end
    return false
end

function BaseCreature:isMoving()

    if self.state == "moving" then
        return true
    end
    return false
end

function BaseCreature:isAttacking()

    if self.state == "attacking" then
        return true
    end
    return false
end

function BaseCreature:update(dt)
    if self._update then
        if self._animation then
            self._animation:update(dt)
        end
        if not game:playersTurn() then
            self:live()
        end
    end
    if self._draw then
        game:drawCreature(self._x,self._y,self.flip,self)
        
    end
end

function BaseCreature:deductEnergy()
    self.energy = self.energy - 1
end

function BaseCreature:increaseHunger()
    self.hunger = self.hunger + 1
    self:deductEnergy()
end

function BaseCreature:_hype()
    if game.debug then
        log(self.name.." is getting hyped.")
    end
    self.hunger = 20
 --   self.energy = 200
end

function BaseCreature:live(dt)
    local state = self.state
    self.tmpSpeed = self.speed
    
    if self:isMoving() then
        self:move()
    elseif self:isFollowing() then
        self:follow()
    elseif self:isAttacking() then
        self:attack()
    elseif self:isHunting() then
        self:hunt()
    elseif self:isSleeping() then
        self:sleep()
    elseif self:isTired() then
        self:tired()
    elseif self:isHungry() then
        self:hungry()
    elseif self:isDead() then
        self:die()
    end
    if self.repeatTurn then
        self.repeatTurn = false
        self:live(dt)
    end
    
    self:deductEnergy()
   -- self:increaseHunger()
    
end

function BaseCreature:die()
    if game.debug then
        log(self.name.." is dying.")
    end
    self._update = false
    self.frame_name = "dead"
    self.hasDied = true
end

function BaseCreature:attack()
    if game.debug then
        log(self.name.." is attacking "..self.target.name)
    end
    self.state = "hunting"
    self:_hype()

end

function BaseCreature:tired()
    if game.debug then
        log(self.name.." is tired.")
    end
    self.state = "sleeping"
    self.repeatTurn = true
end

function BaseCreature:sleep()
    if game.debug then
        log(self.name.." is sleeping.")
    end
    self.energy = self.energy + 100
    if not self:isTired() then
        self.state = "awake"
        self.hunger = 11
        if game.debug then
            log(self.name.." woke up.")
        end
    end
end

function BaseCreature:hungry()
    if game.debug then
        log(self.name.." is hungry ")
    end
 --   self.energy = 40000000000
    self.state = "hunting"
    self.repeatTurm = true
end

function BaseCreature:hunt()
    if game.debug then
        log(self.name.." is hunting")
    end
    self._isHunting = true
    self.target = self.foodTarget
    local moved = self:roam()
    local target = self:canSeeTarget()
    
    if target then
        self.target = target
        self.targetGotten = true
        self.previousState.follow = self.state
        self.state = "following"
        
        if not moved then
            self.repeatTurn = true
        end
    else 
        self.targetGotten = false
    end
    if self.targetGotten and self:targetInAttackRange() then
        self.previousState.attack = self.state
        self.state = "attacking"
    end
end

function BaseCreature:targetInAttackRange(target)
    self.target = target or self.target
    self.targetInRange = nil
    local fov = game.rot.FOV.Precise:new(lightCallback)
    creature = self
    fov:compute(self._x,self._y,self.attackRange,checkTileInRange)
    game:log(game.lume.serialize(self.targetInRange).."yaaaa")
    return self.targetInRange
    
end

function BaseCreature:canSeeTarget()
    if self.targetGotten then
        self.hasTarget = nil
        local fov = game.rot.FOV.Precise:new(trueLightCallback)
        creature = self
        fov:compute(self._x,self._y,self.sight+10,checkTile)
        return self.hasTarget
    else
        self.hasTarget = nil
        local fov = game.rot.FOV.Precise:new(lightCallback)
        creature = self
        fov:compute(self._x,self._y,self.sight,checkTile)
        return self.hasTarget
    end
end

function BaseCreature:follow(target)
    target = target or self.target
    if game.debug then
        log(self.name.." is following "..target.name)
    end
    if  self:targetInAttackRange() then
        self.previousState.attack = self.state
        self.state = "attacking"
    elseif self:canSeeTarget() then
        game:log("following target")
        self.path = self:findPathTo(target)
        self._currentPosInPath = 2
        if not self.path then
            return
        end
        self.previousState.move = self.state
        self.state = "moving"
        self.repeatTurn = true
    else
        self.state = self.previousState.follow
    end

end

function BaseCreature:move()
    if game.debug then
        log(self.name.." is moving.")
    end
    local c = nil
    if self.path then
        c = self.path[self._currentPosInPath]
    end
    if c then
        local x = c.x
        local y = c.y
        local map = game._dungeon
        local skip = nil
        if map[x] and map[x][y] then
            tile = map[x][y]
            log(lume.serialize(lume.keys(tile.creatures)).."heeeeey")
               
            if #lume.keys(tile.creatures) > 0 then tile.isPassable = false
                log("skipping")
                --#to stop by target
                skip = true
                tile.isPassable = false
                self._currentPosInPath = self._currentPosInPath+1
                self.state = self.previousState.move
                end
            if not skip then
            log("not skipping"..lume.serialize(tile.creatures))
                if self.currentTile then
                    self.currentTile.isPassable = true
                    self.currentTile.creatures = {}
                end
                tile.creatures[self.name] = self
                self.currentTile = tile
                tile.isPassable = false
            end
        end
        if not skip then
            if c.x > 0 then
                self.flip = 1
            elseif c.x < 0 then
                self.flip = -1
            end
            log("moving from "..self._x..","..self._y.." to "..lume.serialize(c))
            self._x = c.x --+ self._x
            self._y = c.y --+ self._y
            self._currentPosInPath = self._currentPosInPath+1
            self.tmpSpeed = self.tmpSpeed - 1
            if self.tmpSpeed > 0 then
                self.repeatTurn = true
            end
            if self.previousState.move == "following" then
                if self.behaviors.highInertia == nil then
                    self.state = self.previousState.move
                end
            end
        end
    else
        self.state = self.previousState.move
    end
end

function BaseCreature:findPathTo(target)
    target = target or self.target
    return game:findPath(target,{x=self._x,y=self._y})
    
end

return BaseCreature