ALL_ITEMS = {}
All_Items = ALL_ITEMS
all_items = All_Items
AllItems = all_items
allItems = AllItems

Item = class:extend("Item")

function Item:new(_kwargs)
    local kwargs = _kwargs or self._kwargs
    local t = setmetatable({}, self)
    t.__init__ = self.__init__
    t:__init__(kwargs)
    return t
end

function Item:__init__(kwargs)
    self.isEaten = false
    self.isEquipped = false
    self.isUsed = false
    
    self._kwargs = kwargs or {}
    
    self.isEdible = kwargs.isEdible or false
    self.isPerishable = kwargs.isPerishable or true
    self.isBurnable = kwargs.isBurnable or false
    self.isUsable = kwargs.isUsable or true

    self.name = kwargs.name or "notorious item"
    --Types should fall under:
    ----food
    ----weapon
    ----ammo
    ----projectile
    ----equipment
    --for sorting purposes.
    self.type = kwargs.type or "equipment"
    self.source = game:getSource(kwargs.source or "item.png")

    self.maxCount = kwargs.max or kwargs.maxCount or 10
    
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    self.w = kwargs.w or 16
    self.h = kwargs.h or 16
    self._w = 0
    self._h = 0
    self.animation = nil
    
    self.stats = {
        health = kwargs.health or 10,
        attack = kwargs.attack or 4
    }
    for _,i in pairs(kwargs.stats or {}) do
        self.stats[_] = i
    end
    
    for _,i in pairs(kwargs) do
        self[_] = self[_] or lume.copy(i)
    end
    
    self.effect = kwargs.effect or {}
    self.updatedEffect = kwargs.update or kwargs.updatedEffect or {}
    
    ALL_ITEMS[self.name] = self
    
end

function Item:update(dt)
    self._w,self._h = resizeImage(self.source,self.w,self.h)
    if self.animation then
        self.animation:update(dt)
        if self.animation.done then
            self.animation = nil
        end
    end
    if self.stats.life <=0 then
        if self.isPerishable then
            self.isUsable = false
            self.isEquipable = false
            self.isEdible = false
        end
    end
end

function Item:draw()
    love.graphics.draw(self.source,self.x,self.y,0,self._w,self._h)
end

function Item:eat(user)
    if self.isEdible then
        self.isEaten = true
        self.stats.health = self.stats.life - self.stats.attack
        local data = game:decodeScript(self.effect, user, self)
        self.isEaten = false
        return true
    end
end

function Item:equip(user)
    if self.isEquipable then
        self.isEquipped = true
        self.stats.health = self.stats.life - self.stats.attack
        local data = game:decodeScript(self.effect, user, self)
        self.isEquipped = false
        return true
    end
end
        
function Item:use(user)
    if self.isUsable then
        self.isUsed = true
        self.stats.health = self.stats.life - self.stats.attack
        local data = game:decodeScript(self.effect, user, self)
        self.isUsed = false
        return true
    end
end

--[[
example of an item:
    Bread = Item({
        name = "bread",
        isEdible = true,
        isPerishable = true,
        isBurnable = true,
        life = 3,
        attack = 1, --can be used three times
        effect = {
            actor = "user",
            _if = {"get","self","isEaten","==",true},
            health = {
                "get",
                "user","health","+",
                5
            },
            --act = {"luaCode",...},
            --wait = time to wait before action
        },
        source = "bread.png"
    }
    
    bread1 = Bread:new()
]]--
           
return Item