UNIT_SPEED=0.098
Unit = class:extend("Unit")
--fix enemy's speed limit by applying it(~line 256) and then fix turns...????

function Unit:__init__(kwargs)
    local kwargs = kwargs or {}
    
    self.x = 0
    self.y = 0
    self.w = kwargs.w or 16
    self.h = kwargs.h or 16
    self._w = 1
    self._h = 1
    self.flip = {x=1,y=1}
    self.stats = {
        alias = kwargs.alias or "PLAYER",
        class = kwargs.class or "Borkman",
        
        movementRange = 3,
        movementSpeed = 3,
        
        attackRange = 1,
        healRange = 1,
        
        sight = 14,
        health = 10,
        strength = 10,
        magic = 0,
        --speed of the body and its parts
        speed = 0,
        --flexibility of the body
        agility = 0,
        --number of moves per turn
        speedLimit=2,
        
        intelligence = 3,
        cowardice = 2,
        
        meatpower = 3,
        firepower =  0,
        earthpower = 0,
        windpower = 0,
        waterpower = 0,
        woodpower = 0,
        metalpower = 0,
        
        active = true,
        sleeping = false,
        smell = 2,
        willpower = 1,
        talent = 1,
        energy = 10,
        hunger = -10,
        
        maxItems = 10,
        maxItemCount = 10,
        
        chargeProb = 1,--0.5,
        retreatProb = 0.5,
        
        attacks = {Hit:new(self)},
        morale = 0,
        currentState = "Waking",

    }
    self.bonus = {}
    self.stats.totalHealth = self.stats.health
    
    self.inventory = {}
    self.possibleAttacks = {}
    self.currentAttack = nil
    
    self.itemTypes = {}
    self.itemCount = 0
    self.id = lume.uuid()
    
    self._beforeDrawTrash = {}
    self._afterDrawTrash = {}
    self._drawTrash = {}
    self._updateTrash = {}
    
    self.source = game:getAsset(lume.choice({"player.png","red.png"}))
        
    self.timer = Timer.new()
    self.goto = {x=0,y=0}

end

function Unit:changeClass(class)
    self.stats.oldClass = self.stats.class
    self.stats._oldClass = self.stats._class
    self.stats.class = class
    self:initClass()
end

function Unit:initClass()
    self.stats._class = CLASSES[self.stats.class]:new({unit=self})
end

--take away the rank of commander from a unit
--can have morale impact?
function Unit:takeCommander()
    self.stats.commander = false
    self.stats.morale = self.stats.morale - 1
end

--give the rank of commander to a unit
--can have morale impact?
function Unit:giveCommander()
    self.stats.commander = true
    self.stats.morale = self.stats.morale + 1
end

function Unit:addItemToInventory(item,n)

    local i = self.inventory
    if type(item) == "string" then
        item = ALL_ITEMS[item]
    end
    if #self.itemTypes >= self.stats.maxItems then
        return false
    end
    
    i[item.name] = i[item.name] or (table.insert(self.itemTypes,item.name)) or {}
    if #i[item.name] > item.maxCount then
        return false
    end
    
    local it = (n and item:new()) or item
    
    table.insert(i[item.name],it)
    self.itemCount = self.itemCount+1
    player.heldItems[it.name] = player.heldItems[it.name] + 1
    return true
end

function Unit:removeItemFromInventory(item)

    local i = self.inventory
    local n = item
    
    if type(item) == "string" then
        if i[n] == nil then
            return false
        end
        if #i[n] == 0 then
            lume.remove(self.itemTypes,n)
            i[n] = nil
            return false
        end
        table.remove(i[n],1)
        if #i[n] == 0 then
            lume.remove(self.itemTypes,n)
            i[n] = nil
        end
        self.itemCount = self.itemCount-1
        player.heldItems[n] = player.heldItems[n] - 1
        return true
        
    else
        if n.__id then
            if i[n.name] == nil then
                return false
            end
            if #i[n.name] == 0 then
                lume.remove(self.itemTypes,n)
                i[n.name] = nil
                return false
            end
            lume.remove(i[n.name],n)
            if #i[n.name] == 0 then
                lume.remove(self.itemTypes,n.name)
                i[n.name] = nil
            end
            self.itemCount = self.itemCount-1
            player.heldItems[n.name] = player.heldItems[n.name] - 1
            return true
        end
    end
    return false
end

function Unit:removeItem(...)
    return self:removeItemFromInventory(...)
end

function Unit:addItem(...)
    return self:addItemToInventory(...)
end

function Unit:attackActive()
    return self.currentAttack and self.currentAttack.active
end

function Unit:getAttack()
    return (self.stats.strength-(self.stats.cowardice/2)+self.stats.speed)
end

function Unit:getDamageTo(target)
    local target = target or self.target or Unit:new()
    return self:getAttack()-
        (self.stats.strength-target.stats.strength)-
            (self.stats.speed-target.stats.agility)
end

function Unit:getDamage()
    return self.stats.totalHealth-self.stats.health
end

function Unit:transform()
    --Change class of the unit
end

function Unit:deactivate()
    if self.deactivated then
        return
    end
    self.deactivated = true
    self.stats.active = false
    --self.source = game:getAsset("none.png")
    self.flip.y = -1
    self.path = nil
    if self.selected and not self.isEnemy then
        self.map.player:resetChoice()
        self.selected = nil
    end
end

function Unit:isMyTurn()
    return self.map.turn.data == self.stats.alias
end

function Unit:play()
    self.old_stats = lume.copy(self.stats)
    self.old_tile = self.tile
end

function Unit:replay()
    self.stats = self.old_stats
    self.tile = self.old_tile
    self.played = true

    self.map.turn.data = self.stats.alias
end

function Unit:update(dt)
    assert(self.tile.type~="wall")
    
    self.tile.unit = self
    
    if self.oldCurrentAttack ~= self.currentAttack then
        if self.oldCurrentAttack then
            self.oldCurrentAttack:close()
        end
        self.oldCurrentAttack = self.currentAttack
    end
    
    local trash = self.currentAttack and self.currentAttack:update(dt)
    
    self:trashUpdate(dt)
    
    if self.map.tmpselected==self then
        self.map:colorTile(self.tile,"cyan")
    else
        self.map:uncolorTile(self.tile)
    end
    
    
    if self.timer then
       self.timer:update(dt)
    end
    
    if self.stats.health <= 0 then
        self:deactivate()
    end
    
    if self:isSleeping() then
        self.animation, self.path = nil, nil
    end
    
    self._w,self._h = resizeImage(self.source,self.w,self.h)
    
    if self.shaking and not self.shoken then
        self.map.player.shaking = self.map.player.shaking+1
        self.shoken = true
    elseif not self.shaking and self.shoken then
        self.map.player.shaking = self.map.player.shaking-1
        self.shoken = nil
    end
    
    if not self.animation then
         self.ttg = false
        if self.path then--and self:isMyTurn() then
            local p = self.path[self.currentPosInPath]
            
            if p then
                local tile = self.map.grid[math.floor(p:getX())][math.floor(p:getY())]
                if tile.unit and tile.unit ~= self then
                    log("agaain!")
                    self.map.tmpTarget = self.target
                    self:moveTo(self.path[#self.path]:getPos())
                    if self.path then
                        p = self.path[self.currentPosInPath]
                    --[[elseif not self.isEnemy then
                        --undo turn and give player another chance
                        self:replay()
                        self.map.player:resetChoice()]]
                    else
                    --    error("Noooo path")
                        self.path = nil
                        assert(not self:attackActive())
                        self.done = true
                    end
                    
                end
                local t,fu
                if not p then
             
                    cwarn(tostring(self)..ser(self.path),not self.isEnemy)
                    goto cancel
                end
                    --cwarn(tostring(self)..p:getX()..", "..p:getY(),not self.isEnemy)
                t = self.map.grid[math.floor(p:getX())][math.floor(p:getY())]
                x,y=self.map.map:convertTileToPixel(p:getPos())

                if not self.moves[tostring(t)] then
                    --[[if not self.isEnemy then
                        self:replay()
                        self.map.player:resetChoice()
                        ppp=1
                    end]]
                    goto cancel
                end
                
                fu = function()
                    self.animation.done = true
                --    self.tile.unit = nil
                  --  self.tile = t
                end
                self.animation = Animation({
                     x=x,y=y,
                     d = UNIT_SPEED,
                }):bind("onComplete",fu )
                self.animation:start(self)
                
       
                ::cancel::
                self.currentPosInPath = self.currentPosInPath + 1
                if self.currentPosInPath > self.stats.movementRange+1 then
                    self.currentPosInPath = 0
                end
            else
                self.path = nil
                assert(not self:attackActive())
                self.done = true
            end
            log("00001path: "..self.x..", "..self.y..": "..tostring(self))
        else
        end
    else
        self.animation:update(dt)
        self.ttged = true
        log("000: "..self.x..", "..self.y..": "..tostring(self))
        if self.animation.done then
            self.animation = nil
            log("anim done: "..self:getX()..", "..self:getY()..": "..tostring(self))
        end
        if self:inRange(1) and not self.stats.commander then
            --leave it for ai
            self.path = nil
            WARNING_COLOR = "blue"
            cwarn("ATTACK IS not ACTIVE"..tostring(self),not self:attackActive())
            WARNING_COLOR = "red"
            if not self:attackActive() then
                self.done = true
            end
        end
    end
    
    clog("anim first: "..self:getX()..", "..self:getY()..": "..tostring(self),self.ttged)
      --  log("first: "..self.x..", "..self.y..": "..tostring(self))
    local floor = lume.round

    self.oldTile = self.tile
    
    if self.shaking then
        self.gaga = 1
    end
    
    clog("anim 1: "..self:getX()..", "..self:getY()..": "..tostring(self),self.ttged)
    if not self.shaking and self.gaga then
        self.tile = self.oldTile
        self.x = self.tile._x
        self.y = self.tile._y
        self.gaga = nil
    end
    clog("anim 2: "..self:getX()..", "..self:getY()..": "..tostring(self),self.ttged)


    local x, y = self:getPos()
    local t = self.map.grid[floor(x)] and self.map.grid[floor(x)][floor(y)]
    
    assert(t, "Unit does not have a tile!")
        if self.animation then log("anim: "..x..", "..y..": "..tostring(self)) end
    
    if t ~= self.tile and t then
        self.goto.x,self.goto.y = x*16,y*16
        if self.tile then
            self.tile.unit = nil
        end

        self.tile = t
        self.tile.unit = self
    end
    
    if not self.shaking and not self.animation then
        self.x = self.tile._x
        self.y = self.tile._y
            
    end
    
    clog("anim 33: "..self:getX()..", "..self:getY()..": "..tostring(self),self.ttged)
    if self.selected then
        self.map.colorTiles = {}
        if self.map.tmpselected then
            --self.map.colorTiles[1] = {self.map.tmpselected.tile,"cyan"}
        end
        self.moveID = lume.uuid()
        self.map.fov:compute(x,y,self.stats.movementRange,
            function(x,y,r,v)
                local tile  = self.map.grid[x]and self.map.grid[x][y]
                if not tile then return end
                tile.moveID = self.moveID
                local col = "green"
                if self.stats.alias ~= "PLAYER" then
                    col = "red"
                end
                if tile == self.tile then
                    local ttile = (self.target and (self.target.tile or self.target)) or tile
                    self.map:colorTile(ttile,"orange")
                    self.map:colorTile(tile,col)
                elseif self.map.tmpselected  == self then
                    self.map:colorTile(tile,"red")
                end
            end
        )
    end
    clog("anim end: "..self:getX()..", "..self:getY()..": "..tostring(self),self.ttged)
    
    assert(self.tile, "Unit has no tile")

end

function Unit:getX()
    local x,y = self.map.map:convertPixelToTile(self.x,self.y)
    return math.floor(x)
end

function Unit:getY()
    local x,y = self.map.map:convertPixelToTile(self.x,self.y)
    return math.floor(y)
end

function Unit:getPos()
    return self.map.map:convertPixelToTile(self.x,self.y)
end

function Unit:build(targ)
    local target = targ or self.target
end

function Unit:tame(targ)
    local target = targ or self.target
end

function Unit:heal(targ)
    local target = targ or self.target
    target.stats.health = target.stats.health + self.stats.heal
end

function Unit:healTile(targ)
    local target = targ or self.target
    target.stats.health = target.stats.health + self.stats.heal
end

function Unit:attackTile(targ)
    local target = targ or self.target
    target.stats.health = target.stats.health + self.stats.heal
end

function Unit:inRange(range,target,pass)
    local target = target or self.target
    if target == self or not target or target.isTile then
        return false
    end
    local targetInRange = false
    
    ((pass and self.map.terrainFov) or self.map.fov):compute(self:getX(),self:getY(),range,function(x,y,r,v)
        local tile = self.map.grid[x] and self.map.grid[x][y]
        if tile then
            if tile.unit == target or tile==target then
                targetInRange = true
            end
        end
    end
    )
    
    return targetInRange
end

function Unit:attack(targ)
    --assert((not self.stats.commander and arg.stats.alias~=self.stats.alias) )
    local att = lume.randomchoice(self.possibleAttacks)
    self.target = targ
    att:peform(targ)
    self.currentAttack = att
end

function Unit:inAttackRange(target,x,y)
    local x = x or self:getX()
    local y = y or self:getY()
    self.target = target or self.target
    local target = self.target
    local targetInRange = false
    local t = self.map.grid[x] and self.map.grid[x][y]
    self.possibleAttacks = {}
    
    if not t then return end
    if not self.target then return end

    for _,att in lume.iter(self.stats.attacks) do
        local attRange = att.range
        local inAttRange = false
        if att:canUse(self.target) then
            self.map.fov:compute(self:getX(),self:getY(),attRange,function(x,y,r,v)
                local tile = self.map.grid[x] and self.map.grid[x][y]
                if tile then
                    if tile.unit == self.target then
                        targetInRange = true
                        inAttRange = true
                    end
                end
            end
            )
            if inAttRange then
                table.insert(self.possibleAttacks,att)
            end
        end
    end
    
    return targetInRange
    
end

function Unit:canAttack(target)
    local target = target or self.target
    if target == self then
        return
    end
    local r = self:inAttackRange(target)
    return r
end

function Unit:tooCloseTo(target)
    return false--self:inRange(1)
end

function Unit:tooFarFrom(target)
    return true--not self:inRange(1)
end

function Unit:chargeProb()
    return self.stats.chargeProb
end

function Unit:retreatProb()
    return self.stats.retreatProb
end

function Unit:isTired()
    return self.stats.energy<2
end

function Unit:isHungry()
    return self.stats.hunger>4
end

function Unit:isAlive(other)
    return (o and o.stats.active) or self.stats.active
end

function Unit:isSleeping()
    return self.stats.sleeping
end

function Unit:sleep()
    self.stats.sleeping = true
end

function Unit:getTargets()
    targets = {} k= {}
    self.map.fov:compute(self:getX(),self:getY(),self.stats.sight,
    function(x,y,r,v) table.insert(k,0)
        local tile = self.map.grid[x] and self.map.grid[x][y] 
        if tile then-- self.map:colorTile(tile,"cyan")-- log("yes")
            if tile.unit and tile.unit.stats.active and tile.unit.stats.alias~=self.stats.alias then --tile.unit:isMyTurn() then
                table.insert(targets,tile.unit)
                if not tile.unit.stats then error(tile.unit.__class) end
            end
        end
    end 
    ) 
    return targets
end

function Unit:getObservable(callback)
    local func = callback or NULL
    local tiles = {}
    local val = nil
    
    val = self.map.fov:compute(self:getX(),self:getY(),self.stats.sight,
    function(x,y,r,v)
        local tile = self.map.grid[x] and self.map.grid[x][y]
        assert(tile)
        table.insert(tiles,tile)
        local d = func(tile)
        if d then val = d return d end
    end,true)
    return val or tiles
end

function Unit:getObservables(...)
    return self:getObservable(...)
end

function Unit:notAllied(n)
    return self.stats.alias ~= n.alias
end

function Unit:isAlly(n)
    return not self:notAllied(n)
end

function Unit:getTarget()
    local targs = self:getTargets()
    local weight = nil
    local totalWeigth = nil
    local enemy = nil
    local extra = {}
    local ignore = {}
    local notAllied = function(...) return self:notAllied(...) end
    local unit = self
    if #targs>0 then
        for u,target in lume.iter(targs) do
            target.targeted = target.targeted or 0
            if not target.stats.active then
                goto continue
            end
            extra.h = 0 
            extra.l = 0
            local dis,att,get,def,inAttRange,willAtt,l,h,w,canAttack,targed
            if ignore[tostring(target)] then
                goto continue --for now...(healers, etc)
            end
            
            local dis = lume.distance(target:getX(), target:getY(),unit:getX(),unit:getY())
            local att = unit:getDamage(target)--target.stats.health - unit:getDamage(target)
            local get = targets[tostring(unit)] or 0
    
            --something about enemies in attack range
            --fix values later
            
            local def =  target:getDamage(unit)--unit.stats.health - target:getDamage(unit)
            local inAttRange = unit:inAttackRange(target) and 1 or 0
            local willAtt = 0--unit:inAttackRange(target,) and 1 or 0
            if dis < 2 then
              --  extra.h = extra.h + 10000*(1/dis)
                log("too closeeee!!!")
            end
            extra.h = extra.h+1000*(1/dis)
            local canAttack = inAttRange

            for i,xx in lume.iter(getRound(2)) do
                local x = xx.x+target:getX()
                local y = xx.y+target:getY()
                enem = unit.map.grid[x] and unit.map.grid[x][y]
                if enem and enem.unit and not notAllied(enem.unit) then
                    extra.l = extra.l+enem.unit:getDamage(unit)
                elseif enem and enem.unit and notAllied(enem.unit) then
                    log(inspect(enem.unit,2))
                    extra.h = extra.h+enem.unit:getDamage()
                end
            end
            
            targed = (target.targeted+1)*100
            local l =(targed+dis+att+get+extra.l)+1
            local h = (willAtt +inAttRange+def+1+unit.stats.movementRange+extra.h)+1
            local w = h/l
            assert(h>0 and l>0,"Bad calculation in ai")
            if eneffrmy then
                --floored so target won't keep changing
                if floor(w) <= floor(weight) then
                    log(w.." not better than unit's "..weight)
                    goto continue
                end
            end
            
            if not weight or w > weight then
                weight = w
                totalWeight = w
                enemy = target
                enemy.targeted = enemy.targeted+1
            end
            ::continue::
        end
    end
   -- if #targs~=0 and not self.isEnemy then warn("vudscack") end
    return enemy
end

function Unit:draw()
    self:trashBeforeDraw()
    local sx,sy = 0,0
    if self.flip.y == -1 then
        sy = self.h
    end
    if self.flip.x == -1 then
        sx = self.w
    end
    
    love.graphics.draw(
        self.source,self.x+sx,self.y+sy,
        0,
        self._w*self.flip.x, self._h*self.flip.y
    )
    local trash = self.currentAttack and self.currentAttack:draw()
    self:trashDraw()
    self:trashAfterDraw()
end

function Unit:moveTo(x,y,convert)
    log("start pathfinding")
    self.map.tmpUnit = self
    self.map.tmpTarget = self.target

    local floor = math.floor
    self.map.c_unit = self
    if type(x)=="table" then log(x._x) log(x.getPos)
        convert = false--y
        if x.getPos then
            x,y = x:getPos()
        else
            x,y = x._x,x._y
        end
    else
        log("to "..x..", "..y.." at "..tostring(self))
    end
    if convert then
        x, y = self.map.map:convertPixelToTile(x,y)
    end
    local _x,_y = x,y
    local mx,my = self:getPos()
    if self:inRange(1,self.map.grid[floor(x)][floor(y)].unit,true) then
        log("returning due to in range")
        self.path = nil
        return
    end
    assert(self.map.grid[floor(mx)][floor(my)],"No tile")
    self.moves = {}
    
    self.map.terrainFov:compute(mx,my,self.stats.movementRange,function(x,y)
        local tile = self.map.grid[x] and self.map.grid[x][y]
        if tile then
            self.moves[tostring(tile)] = tile
        end
    end
    )
    
    self.path = self.map:findPathTo(self,x,y)
    self.map:colorTile(self.tile,"blue")
    self.currentPosInPath = 1
    --cwarn(tostring(self)..ser(self.path),not self.isEnemy)
    log("!!!!Returning")
    if self.map.isRogu55e then--???
        if self.path then
            self.path = lume.slice(self.path,1,self.stats.movementSpeed)
        end
    end
    return self.path
end

function Unit:moveAwayFrom(x,y,convert)
    self.map.tmpUnit = self
    local floor = math.floor
    self.map.c_unit = self
    if type(x)=="table" then
        convert = false--y
        if x.getPos then
            x,y = x:getPos()
        else
            x,y = x._x,x._y
        end
    end

    if convert then
        x, y = self.map.map:convertPixelToTile(x,y)
    end
    local _x,_y = x,y
    local mx,my = self:getPos()
    assert(self.map.grid[floor(mx)][floor(my)],"No tile")
    self.moves = {}
    local maxDis = 0
    x = nil
    self.map.terrainFov:compute(mx,my,self.stats.movementRange,function(xx,yy)
        local tile = self.map.grid[xx] and self.map.grid[xx][yy]
        if tile then
            self.moves[tostring(tile)] = tile
        end
        local dis = lume.distance(xx,yy,_x,_y)
        if dis>maxDis then
            maxDis = dis
            x=xx
            y=yy
        end
    end
    )
    
    if x == nil then
        return false
    end
    
    self.path = self.map:findPathTo(self,x,y)
    self.currentPosInPath = 1
    if self.path then log("RUN AWAAAAY!!") end
    return self.path
end

function Unit:standStill()
end


--These can be used by close function in attacks
--after their not current
function Unit:trashDraw()
    for x,i in lume.iter(self._drawTrash) do
        i()
    end
end

function Unit:trashBeforeDraw()
    for x,i in lume.iter(self._beforeDrawTrash) do
        i()
    end
end

function Unit:trashAfterDraw()
    for x,i in lume.iter(self._afterDrawTrash) do
        i()
    end
end
function Unit:trashUpdate(dt)
    for x,i in lume.iter(self._updateTrash) do
        i(dt)
    end
end

function Unit:drawTrash(func)
    table.insert(self._drawTrash,func)
end

function Unit:afterDrawTrash(func)
    table.insert(self._afterDrawTrash,func)
end

function Unit:beforeDrawTrash()
    table.insert(self._beforeDrawTrash,func)
end

function Unit:updateTrash(func)
    table.insert(self._updateTrash,func)
end

function Unit:removeTrash(func)
    return lume.remove(self._updateTrash,func) or
    lume.remove(self._drawTrash,func) or
    lume.remove(self._afterDrawTrash,func) or
    lume.remove(self._beforeDrawTrash,func)
end

return Unit