--[[
    This is the class that holds the reperesentation of a group of
    warriors on the world map.
]]

local love = love
local lume,math = lume,math
--local store = store
local table,pairs,ipairs = table,pairs,ipairs
local getRound, gooi = getRound, gooi
local function dir(p)
    if p>0 then
        return 1
    elseif p<0 then
        return -1
    else
        return 0
    end
end

local Camp = class:extend("Camp")

function Camp:__init__(kwargs)
    local kwargs = kwargs or {}
    self._kwargs = kwargs
    
    self.x = 0
    self.y = 0
    self.source = game:getSource((kwargs.source or "camp.png"))
    self.w = 16
    self.h = 16
    self._w = 0
    self._h = 0
    self.animation = nil
    self.old = {x=0,y=0}
    self.tile = nil
    self.alias = kwargs.alias or "player"
    self.map = kwargs.map or nil
    
    self.stats = {
        movementRange = 30
    }
    
    self.commander = nil
    self.units = kwargs.unit or {}
    self.isCamp = true
    
    self.sight = 5
    self.shaking = 0
    self.animating = 0
    self.battling = false
    self.overworld = nil
    self.noPath = true
    
    self.changeCommand = self.changeCommander
    
end

function Camp:getBattleReady()
    self.battling = true
end

function Camp:endBattle(won)
    self.battling = true
    GSS(self.overworld)
end

function Camp:changeCommander(unit)
    if self.commander then
        self.commander.stats.commander = false
        self.commander:takeCommander()
    end
    self.commander = unit
    unit:giveCommander()
    return unit
end

function Camp:getCommander()
    if not self.commander then
        self:getRandomCommander()
    end
    return self.commander
end

function Camp:getRandomCommander(uts)
    local units = uts or lume.copy(self.units)
    local u = lume.randomchoice(units)
    if u then
        --could have used "while" but 'meh'
        if not u.stats.active or u==self.commander then
            lume.remove(units,u)
            return self:getRandomCommander(units)
        end
        return self:changeCommander(u)
    end
end

function Camp:draw()
    love.graphics.draw(self.source,self.x,self.y,0,self._w,self._h)
end

function Camp:update(dt)
    self._w,self._h = resizeImage(self.source,self.w,self.h)
    if not self.path then
        if not self.noPath then
            
        
            self.map.offset._x = self.tile.x*self.map.tile.w
            self.map.offset._y = self.tile.y*self.map.tile.h
            self.x = self.tile.ttx*self.map.tile.w
            self.y = self.tile.tty*self.map.tile.h
            self._x = self.tile.ttx*self.map.tile.w
            self._y = self.tile.tty*self.map.tile.h
        end
        self.noPath = true
    end
    if not self.animation then
         
        if self.path then
            self.noPath = false
            local p = self.path[self.currentPosInPath]
            
            if p then
                local tile = self.map.tiles[math.floor(p:getX())][math.floor(p:getY())]
                if tile.unit and tile.unit ~= self then
                    self:moveTo(self.path[#self.path]:getPos())
                    if self.path then
                        p = self.path[self.currentPosInPath]
                    else
                        self.path = nil
                        self.done = true
                    end
                    
                end
                local t
                if not p then
                    goto cancel
                end
    
                t = self.map.tiles[math.floor(p:getX())][math.floor(p:getY())]
                x,y=self.map:convertTileToPixel(p:getPos())
                
                 self.animation = Animation({
                     x=x,y=y,
                     d = UNIT_SPEED,
                }):bind("onComplete",function()
                    self.animation.done = true
                end )
                self.animation:start(self)
                --self._x = self.x+(dir(x)*UNIT_SPEED)
              --  self._y = self.y+(dir(y)*UNIT_SPEED)
               -- self.map.offset._x = (self.map.offset._x+dir(x)*UNIT_SPEED)
             --   self.map.offset._y = (self.map.offset._y+dir(y)*UNIT_SPEED)
                ::cancel::
                self.currentPosInPath = self.currentPosInPath + 1
                if self.currentPosInPath > self.stats.movementRange+1 then
                    self.currentPosInPath = 0
                end
                self.tile = t
        
            --self.map.offset._x = self.tile.x*self.map.tile.w
           -- self.map.offset._y = self.tile.y*self.map.tile.h
                self.map.offset.x = t.chunkx
                self.map.offset.y = t.chunky
            else
                self.path = nil
                self.done = true
            end
        else
        end
    else
        self.animation:update(dt)
        if self.animation.done then
            self.animation = nil
            self.x = self.tile.ttx*self.map.tile.w
            self.y = self.tile.tty*self.map.tile.h
            self.map:teleportPlayer(self.tile.ttx,self.tile.tty)
        end
    end
    
    if self.commander then
        if not self.commander.stats.active then
            self:getRandomCommander()
        end
    end
end

function Camp:zoomIn()
    self.map:scaleMap(20)
    self.map:update(0,1)
    local units = {
        units = {}, enemies = {}, w=20,h=20
    }
    for u = 1,15 do
        table.insert(units.units,Unit:new())
    end
    
    self.camp = ZoomedCamp:new(units)
    self.camp.parent = self
    self.map:makeBabyMap(10,self.camp)
    local continue = self.camp:placeUnits()
    if continue then
        GSS(self.camp)
        OLD_STATE = self
        return true
    else
        self.camp = nil
        self.map:scaleMap(1/20)
        return false
    end
    
end

function Camp:getAdgacentCamps(alias)
    if not tile then
        return false
    end
    local camps = {}
    for _,tile in self.tile:getAdgacentTiles() do
        if tile.unit then
            if alias then
                if tile.unit.alias == alias then
                    table.insert(camps,tile.unit)
                end
            else
                table.insert(camps,tile.unit)
            end
        end
    
    end
    return camps
end

function Camp:moveTo(x,y)
    if type(x) == "table" then
        local xx = x
        if xx.isTile then
            x = xx.ttx
            y = xx.tty
        else
            x = (xx.tile and xx.tile.ttx) or xx.x
            y = (xx.tile and xx.tile.tty) or xx.y
        end
    end
    
    self.path = self.map:findPathTo(self,x,y)
    self.currentPosInPath = 1
    return self.path
end
        

return Camp