local floor = math.floor

Dungeon = class:extend("Dungeon")

function Dungeon:__init__(kwargs)
    --self.super.__init__(self,kwargs)
    
    self.data = kwargs.data
    self.player = kwargs.player
    self.units = kwargs.units or {}
    
    local data = self.data
    self.totalFloors = data.totalFloors
    if type(self.totalFloors) == "table" then
        self.totalFloors = lume.randomchoice(self.totalFloors)
    end
    
    self.dungeonFloors = {}
    self.currentDungeon = nil
    self.camera = {}
    self.cameraSpeed = 0
    self.cameraMan = {x=0,y=0}
    self._notMine = {}
    
    local set = function(...) return self.paddy:setButtonText(...) end
    self.set = set
    
    self.defaultFloor = {
        roomCount = {3,5},
        roomWidth = {3,10},
        roomHeight = {3,10},
        corridorWidth = {1,3},
        corridorChance = .8,
        type = "Uniform",
        rooms = {
            {
                id = "normalRoom",
                height = {1,10},
                width = {1,10},
                chance = {1,.2},
                count = {3,7},
            }
        },
        sourceDir = "dungeon/",
        sources = {
            wall = "wall.png",
            upstairs = "upstairs.png",
            downstairs = "downstairs.png",
            water = "water.png",
            door = "door.png",
            floor = "floor.png",
            downstairs = "downstairs.png",
            upstairs = "upstairs.png",
        }
    }
    self:load()
end

function Dungeon:setControls()
    local set = self.set
    set("a","menu")
    set("x","zoom out")
    set("y","zoom in")
    set("b","send out")
end

function Dungeon:load()
    self:loadAllFloors()
    self:setCurrentFloor(1)
end

function Dungeon:loadAllFloors()
    for fl = 1,self.totalFloors do
        self:generateDungeonFloor(fl)
    end
end

function Dungeon:setCurrentFloor(num)
    local floor = self.dungeonFloors[num] or self:generateDungeonFloor(num)
    self.currentDungeon = floor
    self.camera = floor.camera
    self.cameraSpeed = floor.cameraSpeed
    self.cameraMan = floor.cameraMan
    self.paddy = floor.paddy
    
    if not self.paddy._setByDungus then
        self:setControls()
        self.paddy._setByDungus = true
    end
    notMine = self._notMine
    for x,i in pairs(floor) do
        self[x] = (notMine[x] and i) or (self[x]) or i
        notMine[x] = true
    end
    
end

function Dungeon:generateDungeonFloor(floor)
    local dung

    if type(floor) == "number" then
        local _fl = floor
        floor = self.data.floors[floor]
        assert(floor,"No floor at ".._fl)
    end
    
    floor.sources = floor.sources or self.data.floors.sources or self.data.sources
    floor.spawner = floor.spawner or self.data.floors.spawner or self.data.spawner
    self.dungeonFloors[#self.dungeonFloors+1] = self:getDungeon(
    {
        data = floor,
        units = self.units,
        player = self.player,
        parent = self,
    })
    
    dung= self.dungeonFloors[#self.dungeonFloors]
    dung.floor = #self.dungeonFloors
    return dung
end

function Dungeon:getDungeon(floor)
    self:fixFloor(floor)
    local d = DungeonFloor:new(floor)
    d:load(floor)
    d.parent = self
    return d
end

function Dungeon:isDown(...)
    return self.currentDungeon and self.currentDungeon:isDown(...)
end

function Dungeon:draw()
    local trash = self.currentDungeon and self.currentDungeon:draw()
 
    gooi.draw()
    trash = self.pop and self.pop:draw()
    self.camera:draw()
end

function Dungeon:update(dt)
    local trash = self.currentDungeon and self.currentDungeon:update(dt)
    trash = self.pop and self.pop:update(dt)
    gooi.update(dt)
    
    local cameraSpeed = self.cameraSpeed
    local m = {x=self.cameraMan.x,y=self.cameraMan.y}

    for _,u in lume.iter(self.units) do
        u:update(dt)
    end
    
    if self:isDown("zoom in") then
        self.camera.scale= self.camera.scale*1.5
        self:rePopUp()
       -- self.cameraSpeed = self.cameraSpeed/1.5
    elseif self:isDown("zoom out") then
        self.camera.scale = self.camera.scale/1.5
        self:rePopUp()
      --  self.cameraSpeed = self.cameraSpeed*1.5
    elseif self:isDown("left") then
        m.x = self.cameraMan.x-self.tileW
        m.moved = true
    elseif self:isDown("right") then
        m.x = self.cameraMan.x+self.tileW
        m.moved = true
    elseif self:isDown("up") then
        m.y = self.cameraMan.y-self.tileH
        m.moved = true
    elseif self:isDown("down") then
        m.y = self.cameraMan.y+self.tileH
        m.moved = true
    end
    self.cameraMan.x = m.x
    self.cameraMan.y = m.y
    
    if m.moved then
        self:rePopUp()
    end
    if m.momved then--not self.cameraMan.animation and m.moved then
        self.cameraMan.animation = Animuation:new({x=m.x,y=m.y,d=cameraSpeed})
        self.cameraMan.animation:start(self.cameraMan)
        self.cameraMan.animation:update(dt)

    end
    local maxZoom = 10
    local minZoom = 0.05
    if self.camera.scale >  maxZoom then
        self.camera.scale = maxZoom-1
        self.cameraMan.animation= nil
    elseif self.camera.scale < minZoom then
        self.camera.scale = minZoom+0.01
    end
end

function Dungeon:mousereleased(x, y, button)
    self.currentDungeon:mousereleased(x,y,button)
    local trash = self.pop and self.pop:mousereleased(x,y,button)
    gooi.released()
end

function Dungeon:mousepressed(x, y, button)
    self.currentDungeon:mousepressed(x,y,button)
    local trash = self.pop and self.pop:mousepressed(x,y,button)
    gooi.pressed()
end

function Dungeon:getTileWithPos(...)
    return self.currentDungeon and self.currentDungeon:getTileWithPos(...)
end

function Dungeon:touchreleased(id, x, y, dx, dy, pressure)
    self.currentDungeon:touchreleased(id,x,y,dx,dy,pressure)
    if 1 then return end --????
    local trash = self.pop and self.pop:touchreleased(id,x,y,dx,dy,pressure)
    if trash then
        self.buttonPressed = true
    end
    if not self.buttonPressed then
        ttx = x
        tty = y
    else
        self.buttonPressed = false
        return
    end
    x,y = self.camera:toWorldCoords(x,y)
   
    local tile = self:getTileWithPos(x,y)
    local sx,sy = self.map:convertPixelToTile(x,y)
    sx = math.floor(sx)
    sy = math.floor(sy) 
   sprint2 = sx..", "..sy
    if tile then
        local now = nil
       -- table.insert(self.colorTiles,{tile,"red"})
      -- self:colorTile(tile,"red")
       sprint2 = sprint2.."yah"
   --     self.cameraMan = {x=x,y=y}
   
        local target = tile.unit
        self.tmpselected = target
        local unit = self.player.selected
        if tile.unit then
            self.tmpselected = tile.unit
            self:popUp(1,tile.unit)
        end
    end
end

function Dungeon:touchpressed(id, x, y, dx, dy, pressure)
    self.currentDungeon:touchpressed(id,x,y,dx,dy,pressure)
end

function Dungeon:touchmoved(id, x, y, dx, dy, pressure)
    self.currentDungeon:touchmoved(id,x,y,dx,dy,pressure)
end

function Dungeon:popUp(num, obj)
    self._popUp = {num,obj}
    if self.pop then
        self.pop:destroy()
    end
    log(obj.x..", "..obj.y)
    local sw = 8
    local sh = 2
    local xx, yy = self.camera:toCameraCoords(obj.x,obj.y)
    log(xx..",, "..yy)
   -- local yy= 16--00--yy-100
    yy = yy-(obj.h*5)
    if num==1 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"EQUIP","FEED","ITEMS"},
            onRelease=function(t)
                if t=="ITEMS" then
                    self.pop:destroy()
                    self:popUp(3,obj)
                elseif t=="EQUIP" then
                    self.pop:destroy()
                    self:popUp(2,obj)
                elseif t=="FEED" then
                    self.pop:destroy()
                    self:popUp(2,obj)
                    
                end
            end
        })
        self.pop = pop
        
    elseif num==2 then
        local data = player.itemTypes
        local i = obj.inventory
        
        for name, list in ipairs({}) do--player:iterInventory() do
            if ALL_ITEMS[name].isEdible then
                table.insert(data,name)
            end
        end
        table.sort(data, function(a, b) return i[a] end)
        
        local pop = GiverTaker:new({
            data = data,
            x=xx,
            y=yy,
            w=obj.w*sw+20,
            h=obj.h*sh+20,
            inventory = obj.inventory,
            onRelease = function(val,t)
                local lim = player.heldItems[t]
                if val<0 then
                    if self.dont then
                        self.dont = nil
                        return
                    end
                    player:removeItemFromUnit(t,obj)
                elseif val>0 then
                    if lim >= #player.inventory[t] then
                        self.pop.dont = true
                        self.pop.spinner:changeValue(val*-1)
                        return false
                    end
                    player:addItemToUnit(t,obj)
                end
             --   self.pop:destroy()
               -- self:popUp(2,obj)
            end
        })
        self.pop = pop
    end
end

function Dungeon:rePopUp()
    self.camera:update(0.1)
    if not self.pop then
        return
    end
    self.pop:destroy()
    self:popUp(self._popUp[1],self._popUp[2])
end

function Dungeon:textinput(text)
  gooi.textinput(text)
end

function Dungeon:keypressed(key, scancode, isrepeat)
  self.currentDungeon:keypressed(key,scancode,isrepeat)
  gooi.keypressed(key, scancode, isrepeat)
  if key == "escape" then
    p:close()
    love.event.quit()
  end
end

function Dungeon:keyreleased(key, scancode)
  self.currentDungeon:keyreleased(key,scancode)
  gooi.keyreleased(key, scancode)
end

function Dungeon:fixFloor(floor)
    for k,v in pairs(self.defaultFloor) do
        floor[k] = floor[k] or v
        if k == "sources" then
            for x,i in pairs(self.defaultFloor.sources) do
                floor[k][x] = floor[k][x] or i
            end
        end
    end
    return floor
end

return Dungeon