
DungeonFloor = GridMap:extend("DungeonFloor")

function DungeonFloor:__init__(kwargs)
    self:callSuper("__init__",kwargs)

    self.isRogoue = true
    self._data = kwargs.data
    --debugging--???
    floor=self
    
    self.player = kwargs.player
    self.parent = kwargs.parent or kwargs.map
    self.observe = true
    
    self.units = kwargs.units
    self.data = self._data
    local floor = self._data
    
    self.totalIntensity = floor.intensity
    self.intensity = 0
    
    self.creatures = {}
    self.types = {
        [0]="floor",
        [1]="wall",
    }
    
    self.upstairs = nil
    self.downstairs = nil

    --contains all rooms by id
    self.rooms = {}
    --contains all tiles in room by id of the room
    self.tilesInRooms = {}

    floor.w = floor.w or floor.width
    floor.h = floor.h or floor.height
    floor.height = nil
    floor.width = nil
    if type(floor.w) == "table" then floor.w = lume.randomchoice(floor.w) end
    if type(floor.h) == "table" then floor.h = lume.randomchoice(floor.h) end
    
    self._dungeon = rot.Map[self._data.type](floor.w,floor.h,floor)
    self._dungeon:create(function(...)
        return self:buildDungeon(...)
    end)
    self:assignRoomIDs()
    self:getUpstairs()
    self:getDownstairs()
    
    self:placeUnits()
    
    self.stateMachine = StateMachine:new({
        map = self
    })
    self.turn = Toggle:new({"ENEMY","PLAYER"})
    self.turn.data = "PLAYER"
    self.turn.value = 0
    self.scheduler = Scheduler:new()
    for x,c in lume.iter(self.allTiles) do
        local x,y = c.x, c.y
        self.grid[x] = self.grid[x] or {}
        self.numGrid[x+1] = self.numGrid[x+1] or {}
        self.grid[x][y] = c
        self.mapGrid[y] = self.mapGrid[y] or {}
        self.mapGrid[y][x] = c
    end
    --[[for x in ume.iter self.grid) do
        for y,t in lume,iter(u) do
            self.mapGrid[y] = self.mapGrid[y] or {}
            self.mapGrid[y][x] = c
            -- log(x..", "..y)
            log("map at y: "..y..", x: "..x)
       end
    end]]
    
    self._grid = JumperGrid(self.mapGrid)
    self.pathfinder = PathFinder(self._grid,"ASTAR",
        function(val) return self:isWalkable(val) end
    )
    self.pathfinder:setMode("ORTHOGONAL")
    self:load()
end

function DungeonFloor:placeUnits()

    local theTiles = lume.copy(self:getRoom(self.upstairs).tiles)
    
    for _,xx in lume.iter(self.units) do
        xx.stats.movementRange = 11--?????
        xx.stats.alias = "PLAYER"
        
        tile = eliminate(theTiles)
        if tile then
            self:placeUnit(xx,tile)
        end
    end
    --local s = math.floor(#self.grid/2)
    --local ss = math.floor(#self.grid[math.floor(#self.grid/2)]/2)
    --local t = self.grid[s][ss]
    --self.cameraMan = {x=t._x,y=t._y}
end

function DungeonFloor:placeUnit(unit, tile)

    local _tile = tile
    unit.x = tile._x
    unit.y = tile._y
    unit.w = _tile.w
    unit.h = _tile.h

    unit.tile = tile
    unit.map = self
    tile.unit = unit
end

function DungeonFloor:spawn(unit, room)
    local unit = unit
    
    if type(unit) == "string" or not (unit.update and unit.draw) then
        local _unit = unit
        unit = getNewCreature(_unit) assert(unit)
        if not unit then
            warn("No such creature ".._unit)
            return false
        end
    end
    tile = self:getEmptyTileInRoom(room)
    if not tile then
        return false
    end
    table.insert(self.creatures,unit)
    unit.stats.alias = "ENEMY"
    self:placeUnit(unit, tile)
    return true
end
    
function DungeonFloor:load(data)
end

function DungeonFloor:buildDungeon(x,y,val)
    local type = self.types[val]
    _t = {
        type = type,
        source = self.data.sources[type],
    }
    self:addTile(_t,x,y)
end

function DungeonFloor:isWalkable(val)
    if val.type == "wall" then return false end
    if val.unit and val.unit ~= self.tmpUnit and val.unit ~= self.tmpTarget then return false end
    if val.collidable then
        return false
    else
        return true
    end
end

--iterate through all units and creatures
function DungeonFloor:allUnits()
    local x = 0
    local nxt = false
    local val
    
    return function()
        x = x+1
        val = self.units[x]
        if not val or nxt then
            nxt = true
            x = 1
            val = self.creatures[x]
        end
        return val
    end
end

function DungeonFloor:update(dt)
    local trash = self.pop and self.pop:update(dt)
    self.super.update(self,dt)
    
    self:getTurn()
    
    for x,i in lume.iter(self.creatures) do--self:allUnits() do
        assert(i.stats.alias~="PLAYER",ser(i.stats.alias))
        i:update(dt)
    end
    self.player:update(dt)
    self.scheduler:update(dt)
    
    
    
    if self.intensity < self.totalIntensity+10 then
        --log("intensity: "..self.intensity.."/"..self.totalIntensity)
        --log("spawning some creatures!")
        local r = self:getRooms()
        assert(#r>0)
        self.player.units = self.units
        self.player.commander = self.player:getCommander()
        
        
        lume.remove(r,self:getRoom(self.player.commander.tile.roomID))
        if #r<0 then
            return
        end
        
        local room = lume.randomchoice(r)
        assert(room,inspect(r,2))
        
        local limit = self.totalIntensity-self.intensity
        local min = lume.min(self.data.spawner)
        local max = lume.max(self.data.spawner)
        max = (max > limit and limit) or max
        
        local intensity = math.random(min,max)
        local monster = self.data.spawner[intensity]
        
        local p = room.properties
        --???if p.spawn then end
        
        if monster then
            local spawn = self:spawn(monster,room)
            if spawn then
                self.intensity = self.intensity+intensity
            end
        end
        --log("done")
        --log("intensity: "..self.intensity.."/"..self.totalIntensity)
    end
end
        
            
--returns false if light can't pass through tile
function DungeonFloor:lightCallback(fov, x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    if tile.unit then return false end
    if tile.type == "wall" then return false end
    return true
end

--returns true as long as the tile exists
function DungeonFloor:tLightCallback(fov, x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    return true
end

function DungeonFloor:onFOVCompute(x, y, range, visibility)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    if tile.type == "wall" then return end
    --self:makeTileDarker(tile)
    
end


function DungeonFloor:draw()
    self.super.draw(self)
    self.camera:attach()
    for _,u in lume.iter(self.units) do
        u:draw()
    end
    for x,i in ipairs(self.creatures) do
        i:draw()
    end
    self.camera:detach()
    local trash = self.pop and self.pop:draw()
    local trash = self.joystick and self.joystick:draw()
    self.camera:draw()
end

function DungeonFloor:doBattleAI(obj)
    if not obj then
        self.currentUnit.done = false
        self.currentUnit.moved = false
        self.currentUnit.selected = true
        self.currentUnit:play()
    end
    log("---peforming statemachine with "..tostring(self.currentUnit).."---")
    self.stateMachine:peform(self.currentUnit)
    log("---done at state "..self.stateMachine:currentState(self.currentUnit).."---")
    
    if obj then
        if not self.currentUnit.attacking then
            assert(not self.currentUnit:attackActive())
            self.currentUnit.done = true
            self.currentUnit.attacking = true
        end
    end

    --self.scheduler = Scheduler:new()
   -- assert(#self.scheduler._schedules<1,"Still got schedule at "..self.currentUnit.stats.alias)
    self.scheduler:schedule(
    function() log(#self.scheduler._schedules..'!') return self.currentUnit.done end,
    function()
        --check if anything is in range to attack
    --   error("...")
        if not self.currentUnit.attacking then log("!!!!!aa")
          if self.stateMachine:currentState(self.currentUnit) =="Battling" then
            log("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            self.currentUnit.moved = true
            self.currentUnit.done = false
            return self:doBattleAI(self.currentUnit) end
        end
    --    log("!!!!!"..self.currentUnit.count)
        self.currentUnit.attacking = false
        self.currentUnit.selected = false
        --reseting current unit
        assert(self.currentUnit.done)
        self.currentUnit.done = false
        self.currentUnit.target = nil
        --give another unit a shot
        lume.remove(self.actingUnits,self.currentUnit)
        warn(self.stateMachine:currentState(self.currentUnit))
        self.currentUnit = self.actingUnits[1]
        if not self.currentUnit and self.turn.data~="PLAYER" then
            warn("toggling out of"..self.turn.data)
            self.turn:toggle()
        elseif self.currentUnit then
         --   warn("nexttttttt")
            self.currentUnit:play()
            self.currentUnit.selected = true
            self:doBattleAI()
        else
            --commander's time
         --   error()
            self.player.ready = true
            self:colorTile(self.player.commander.tile,"red")
        end
    end
    )
  --  log("schedule done")
end

function DungeonFloor:peformPlayerAI()
    self.actingUnits = lume.copy(self.units)
    lume.remove(self.actingUnits,self.player.commander)
    for x,g in ipairs(self.units) do
        assert(not g.attacking) g.count=x
        g.attacking = nil
        if not g.stats.active then
            lume.remove(self.actingUnits,g)
        end
    end
    self.currentUnit = self.actingUnits[1]
    if not self.currentUnit then
        if not self.player.commander or not self.player.commander.stats.active then
            self.set("...","exit")
            self.player.won = false
            return self.turn:toggle()
        else
            self.player.ready = true
            return
        end
    end
    self:doBattleAI()
end

function DungeonFloor:peformAI(units)
    self.actingUnits = lume.copy(units)
    for x,g in ipairs(units) do
        assert(not g.attacking)
        assert(g.stats.alias ~="PLAYER")
        g.attacking = nil
        if not g.stats.active then
            lume.remove(self.actingUnits,g)
            warn("not active enemy")
        end
    end
    warn(#self.actingUnits.."!!!?")
    self.currentUnit = self.actingUnits[1]
    if not self.currentUnit then
        return self.turn:toggle()
    end
    self:doBattleAI()
end

function DungeonFloor:getTurn(dt)
    if self.turn.data == "ENEMY" then-- and
        if self.PLAYER then
         --   self.turn.value = self.turn.value+1
            self:loadTurn()
            log("\nENEMY TURN\n")
            self:peformAI(self.creatures)
        end
        self.PLAYER=nil
    elseif not self.PLAYER and self.turn.data == "PLAYER" then
        log("\nPLAYER TURN\n")
        self.turn.value = self.turn.value+1
    --    warn()
        self:loadTurn()
        self.player.ready = false
        self.player.moved = false
        self.PLAYER=1
        self:peformPlayerAI()
    end
end

function DungeonFloor:getRoom(id)
    --tile is a corridor--????
    if not id then
        return
    end
    if type(id) == "table" and id.isRoom then return id end
    if type(id) == "table" then
        id = id.roomID
    end
    return self.rooms[id]
end

function DungeonFloor:isWallCallback(x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if tile then
        if tile.type == "wall" then
            return true
        else
            return false
        end
    end
end

function DungeonFloor:addDoorToMap(x,y)
    tile = self.grid[x][y]
    tile.source = "door.png"
    tile.type  = "door"
    tile.isPassable = true
end

function DungeonFloor:lockRoom(dungeon)
    local corridors = self._dungeon:getCorridors()
end

function DungeonFloor:addDoors(room,n)
    local n = n or 1
    for g = 1,n do
        room:addDoor(x,y)
    end
    room:getDoors(function(...) return self:addDoorToMap(...) end )
end

function DungeonFloor:addAllDoorsToMap()
    for _,t in pairs(self._dungeon:getRooms()) do
        t:getDoors(function(...) return self:addDoorToMap(...) end )
    end
    --room:addDoors(isWallCallback)--(x,y,return true?)
end

function DungeonFloor:addTileToRoom(tile,roomID)
    if tile._inserted_in_room then
        return
    end
    local room = (type(roomID) == "table" and roomID) or self:getRoom(roomID)
    local roomID = (type(roomID) == "table" and roomID.id) or roomID
    tile.roomID = roomID
    --tile.color = getColor("red")
    table.insert(room.tiles,tile)
    if tile.type == "wall" then
        table.insert(room.walls,tile)
    else
        table.insert(room.spaces,tile)
    end
    tile._inserted_in_room = true
end

function DungeonFloor:_getTile(x,y,room,noUnits)
    local dung = self.grid[x][y]
    if dung.units and noUnits then
        dung = nil
    else
        self:addTileToRoom(dung,room)
    end
    return dung
end

function DungeonFloor:getRoomTiles(room, noUnits)
    if #room.walls>0 then return {floors = room.spaces, walls = room.walls} end
    
    --check Bottom should actually be top(in kivy coords) else switch
    local spaces = {}
    local dungeon = self.grid
    local walls = {}
    
    top = room:getTop()
    bottom = room:getBottom()
    left = room:getLeft()
    right = room:getRight()
    for t = top,bottom do
        for r = left,right do
            local dung = self:_getTile(r,t,room,noUnits)
            --self:colorTile(dung,"red")
            if dung then
                if dung.type == "wall" then
                    table.insert(walls,{x=r,y=t,[1]=r,[2]=t})
                else
                    table.insert(spaces,{x=r,y=t,[1]=r,[2]=t})
                end
            end
            
            if t==bottom then
                local dung = self:_getTile(r,t-1,room,noUnits)
                if dung then
                    if dung.type == "wall" then
                        table.insert(walls,{x=r,y=t-1,[1]=r,[2]=t-1})
                    else
                        table.insert(spaces,{x=r,y=t-1,[1]=r,[2]=t-1})
                    end
                end
            end
            if t==top then
                local dung = self:_getTile(r,t+1,room,noUnits)
                if dung then
                    if dung.type == "wall" then
                        table.insert(walls,{x=r,y=t+1,[1]=r,[2]=t+1})
                    else
                        table.insert(spaces,{x=r,y=t+1,[1]=r,[2]=t+1})
                    end
                end
                
            end
            if t==left then
                local dung = self:_getTile(r-1,t,room,noUnits)
                if dung then
                     if dung.type == "wall" then
                        table.insert(walls,{x=r-1,y=t,[1]=r-1,[2]=t})
                    else
                        table.insert(spaces,{x=r-1,y=t,[1]=r-1,[2]=t})
                    end
                end
                
            end
            if t==right then
                local dung = self:_getTile(r+1,t,room,noUnits)
                if dung then
                    if dung.type == "wall" then
                         table.insert(walls,{x=r+1,y=t,[1]=r+1,[2]=t})
                    else
                        table.insert(spaces,{x=r+1,y=t,[1]=r+1,[2]=t})
                    end
                end
                
            end
        end
    end
    return {floors=spaces,walls=walls}
end

function DungeonFloor:getEmptyRoom()
    local count = 0
    local _repeat = true
    while _repeat do
        if count>100 then
            break
        end
        count = count + 1
        _repeat = false
        local r = self:getRandomRoom()
        check = self:getRoomTiles(r)
        for _,x in ipairs(check.floors) do
            if self.grid[x.x][x.y].unit then
                _repeat = true
                break
            end
        end
    end
    return check
end

function DungeonFloor:assignRoomIDs()
    local r = self._dungeon:getRooms()
    for x,room in ipairs(r) do
        room.id = lume.uuid()
        local rand = math.random()*10
        
        for x,ro in pairs(self.data.rooms) do
            if rand>ro[1] and rand<ro[2] then
                room.properties = lume.copy(self.parent.data[x])
                break--???
            end
        end
        
        self.rooms[room.id] = room
        room.tiles = {}
        room.walls = {}
        room.spaces = {}
        room.isRoom = true
        self:getRoomTiles(room)
    end
end

function DungeonFloor:getRooms()
    return self._dungeon:getRooms()
end 

function DungeonFloor:getRandomRoom(rooms)
    local r
    if not rooms then
        r = self._dungeon:getRooms()
    else
        r = rooms
    end
    rr = lume.randomchoice(r)
    return rr
end

function DungeonFloor:getRandomSpaceInRoom(room)
    local room = room or self:getRandomRoom()
    spaces = self:getRoomTiles(room)
    return lume.randomchoice(spaces.floors)
end

function DungeonFloor:getEmptyTilesInRoom(room)
    local room = room or self:getRandomRoom()
    spaces = self:getRoomTiles(room,true)
    return spaces.floors
end

function DungeonFloor:getEmptyTileInRoom(room)
    local tiles = self:getEmptyTilesInRoom(room)
    if #tiles < 1 then
        error("Tiles to low")
        return false
    end
    r = lume.randomchoice(tiles)
    return self.grid[r.x][r.y]
end
        
function DungeonFloor:getUpstairs()

    if self.upstairs then
        return self.upstairs
    end
    
    local room = self:getRandomRoom()
    sp = self:getRandomSpaceInRoom(room)
    local tile = self.grid[sp.x][sp.y]
    self.upstairs = tile
    tile.isUpstairs = true
    tile.source = self.data.sources["upstairs"]
    return tile
end

function DungeonFloor:getDownstairs()
    
    local room = self:getRandomRoom()
    
    if room.id == self.upstairs.roomID then
        return self:getDownstairs()
    end
    
    sp = self:getRandomSpaceInRoom(room)
    local tile = self.grid[sp.x][sp.y]
    self.downstairs = tile
    tile.isDownstairs = true
    tile.source = self.data.sources["downstairs"]
    return tile
end

function DungeonFloor:mousereleased(x, y, button)
    self.super.mousereleased(self,x,y,button)
    local trash = self.pop and self.pop:mousereleased(x,y,button)
    local trash = self.joystick and self.joystick:released()
    gooi.released()
end

function DungeonFloor:mousepressed(x, y, button)
    local x,y = love.mouse.getPosition() warn(x..",!!! "..y)
    self.super.mousepressed(self,x,y,button)
    local trash = self.pop and self.pop:mousepressed(x,y,button)
    local trash = self.joystick and self.joystick:pressed()
    gooi.pressed()
end

function DungeonFloor:touchpressed(id, x, y, dx, dy, pressure)
    self.super.touchpressed(self,id,x,y,dx,dy,pressure)
    local trash = self.pop and self.pop:touchpressed(id,x,y,dx,dy,pressure)
end

function DungeonFloor:touchreleased(id, x, y, dx, dy, pressure)
    self.super.touchreleased(self,id,x,y,dx,dy,pressure)
    local trash = self.pop and self.pop:touchreleased(id,x,y,dx,dy,pressure)
    if trash then
        self.buttonPressed = true
    end
    if self.turn.data ~= "PLAYER" then
        return
    end
    if not self.buttonPressed then
        ttx = x
        tty = y
    else
        self.buttonPressed = false
        return
    end
    x,y = self.camera:toWorldCoords(x,y)
   
    local tile = self:getTileWithPos(x,y)
            local sx,sy = self.map:convertPixelToTile(x,y)
            sx = math.floor(sx)
            sy = math.floor(sy) sprint2 = sx..", "..sy
    if tile then
        if self.unitDisplay then
            self.unitDisplay:destroy()
            self.unitDisplay = nil
        end
        if 1 then
            self:observeNoise(tile,3)
        end
        local now = nil
       -- table.insert(self.colorTiles,{tile,"red"})
       self:colorTile(tile,"darkblue")
       tile.color = getColor("darkblue")
       sprint2 = sprint2.."yah"
   --     self.cameraMan = {x=x,y=y}
   
        local target = tile.unit
        self.tmpselected = target
        local unit = self.player.selected
        if tile.unit then
            assert(tile.unit.tile==tile)
            local st = tile.unit.stats
            self.unitDisplay = UnitDisplay:new({
                unit = tile.unit,
                health = st.health,
                strength = st.strength,
                agility = st.agility,
                alias = st.alias,
                name = st.name,
                job = st.job,
                species = st.species,
                --target = ...,
                --favAttack = ...,
                --falvTeammate = ...,9
            })
            
            self.tmpselected = tile.unit
            if not self.player.selected then
                if target.stats.alias ~= "PLAYER" then
                    self:popUp(2,target)
                elseif not target.stats.commander then
                    self:popUp(1,target)
                elseif target.stats.commander then
                    self:popUp(3,target)
                end
            elseif self.player.selected and not target.stats.active and
            unit:inRange(1,target) then
                self:popUp(9,target)
            elseif self.player.selected and not self.player.moved then
                if target==unit then
                    --end turn
                    self:popUp(6,unit)
                elseif target and unit:inAttackRange(target) then
                    --attack
                    self:popUp(5,target)
                elseif  target.stats.alias == "PLAYER" then
                    --manager of units?
                    --but for now...
                        local f = self.pop and self.pop:destroy()
                        self.pop = nil
                end
            elseif self.player.selected and self.player.moved then
                if target==unit then
                    --undo
                    self:popUp(8,target)
                elseif target and unit:inAttackRange(target) then
                    --attack
                    self:popUp(5,target)
                end
            end
        else
            self.tmpselected = tile
            if self.player.moved and unit:inAttackRange(tile) then
                --attack a tile?
                self:popUp(7,tile)
        
            elseif self.player.selected and not self.player.moved then
                --move
                self:popUp(4,tile)
            else
                --remove all popUps?
                if self.pop then
                    self.pop:destroy()
                    self.pop = nil
                end
                --self:popUp(1,tile)
            end
        end
    end
end

function DungeonFloor:moveTo(tile)
    --move player.selected to a tile
    self.buttonPressed = true
    assert(tile~=self.player.commander)
    local sx,sy = tile:getX(),tile:getY()
    local s = self.player.selected
    s.oldx, s.oldy = s:getPos()
    if  self.player.selected.moveID == tile.moveID 
      and 
      self.player.selected.tile ~= tile then
        sprint2 = sx..", "..sy..": "..tostring(tile.unit)
        --self.tmpselected = tile.unit
        if tile ~= self.player.selected.tile then
            log("Staaararararat")
            local moved = self.player.selected:moveTo(sx,sy)
            if not moved then
                return
            end
            self.player.selected._goal = {sx,sy}
            self.player.moved = self.player.selected
            --self.turn:toggle()
            self.colorTiles = {}
            if self.pop then
                self.pop:destroy()
            end
            self.pop = nil
        end
        
        self.colorTiles = {}
    end
    -- self.colorTiles = {}
end

function DungeonFloor:endturn()
    
        if self.pop then
            self.pop:destroy()
        end
        self.pop = nil
        self.buttonPressed = true
        self.colorTiles = {}
        local s = self.player.selected
        if s then
            s.selected = false
        end
        self.player.selected = nil
        self.player.moved = false
        self.player.ready = false
        self.turn:toggle()
        --self.enemyAI.acted = false
end

function DungeonFloor:popUp(num, obj)
    if not self.player.ready and
         num ~=1 then
         return
    end
    self._popUp = {num,obj}
    self.playerMoved = true
    if self.pop then
        self.pop:destroy()
    end
    local sh = 2
    local sw = 8--16
--elf.camera.scale
    gooi.removeComponent(self.pop)
    local g= self.grid[1][1]
    local fx,fy = self.map:convertPixelToTile(-10,30)
    local x,y = obj.x,obj.y
    if obj.isTile then
        x,y = obj._x,obj._y
    end
    local xx, yy =self.camera:toCameraCoords(x,y)
   -- local yy= 16--00--yy-100
    yy = yy-(obj.h*5)
    
    local select = function(obj)
        self.pop:destroy()
        self.buttonPressed = true
        self.colorTiles = {}
        if self.player.selected then
            self.player.selected.selected = false
        end
        obj:play()
        self.player.selected = obj
        obj.selected = true
        self.player.selected.moveID = lume.uuid()
        local sx,sy = self.map:convertPixelToTile(obj.tile.x,obj.tile.y)
        sx = math.floor(sx)
        sy = math.floor(sy)
        self.fov:compute(sx,sy,self.player.selected.stats.movementRange,
            function(x,y)
                local tile  = self.grid[x] and self.grid[x][y]
                if not tile then return end
                tile.moveID = self.player.selected.moveID
                --self:colorTile(tile,"red")
            end
        )
    end

    local endturn = function()
    
        if self.pop then
            self.pop:destroy()
        end
        self.pop = nil
        self.buttonPressed = true
        self.colorTiles = {}
        local s = self.player.selected
        if s then
            s.selected = false
        end
        self.player.selected = nil
        self.player.moved = false
        self.player.ready = false
        self.turn.data = "ENEMY"
    end
    self.endturn = endturn
    
    if num == 1 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"MAKE COMMANDER","END TURN"},
            onRelease=function(t)
                if  t=="END TURN" then
                    endturn()
                elseif t=="MAKE COMMANDER" then
                    self.player:changeCommander(self.tmpselected)
                    self.player.moved = false
                    self.pop:destroy()
                    self.pop = nil
                end
            end
    
        })
        self.pop = pop
    
    elseif num == 2 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"END TURN"},
            onRelease=function(t) if  t=="END TURN" then endturn() end end
    
        })
        self.pop = pop

    elseif num == 3 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"SELECT","END TURN"},
            onRelease=function(t)
                if t=="SELECT" then
                    select(obj)
                elseif t=="END TURN" then
                    endturn()
                end
            end
        })
        self.pop = pop
        
    elseif num == 4 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"MOVE HERE","END TURN"},
            onRelease=function(t)
                if t=="MOVE HERE" then
                    self:moveTo(obj)--self.tmpselected)
                elseif t=="END TURN" then
                    endturn()
                end
            end
        })
        self.pop = pop
        
    elseif num == 5 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"ATTACK"},
            onRelease=function(t)
                if t=="ATTACK" then
                    local s = self.player.selected
                    s.target = obj
                    if s:inAttackRange(obj) then
                        s:attack(obj)
                    end
                end
            end
        })
        self.pop = pop
        
    elseif num == 6 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"BACK","END TURN"},
            onRelease=function(t)
                if  t=="END TURN" then
                    endturn()
                elseif t=="BACK" then
                    self.player.selected.selected = false
                    self.player.selected = nil
                    self.pop:destroy()
                    self.pop = nil
                end
            end
    
        })
        self.pop = pop
        
    elseif num == 7 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"ATTACK"},
            onRelease=function(t)
                if t=="ATTACK" then
                    local s = self.player.selected
                    s.target = obj
                    if s:inAttackRange(obj) then
                        s:attackTile(obj)
                    end
                end
            end
        })
        self.pop = pop
        
    elseif num == 8 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"UNDO","END TURN"},
            onRelease=function(t)
                if t=="UNDO" then
                    self.colorTiles = {}
                    local ob = self.player.moved
                    if not ob then
                        log("Can't undo!")
                        return
                    end
                    ob:moveTo(ob.oldx,ob.oldy)
                    self.buttonPressed = true
                    local s = self.player.selected
                    if s then
                        s.selected = false
                    end
                    self.player.moved = false
                    --self.player.selected = nil
                    self.pop:destroy()
                elseif t=="END TURN" then
                    endturn()
                end
            end
        })
        self.pop = pop
        
    elseif num == 9 then
        local pop = PopUp:new({
            x=xx,
            y=yy,
            w=obj.w*sw,
            h=obj.h*sh,
            data={"LOOT","PUSH"},
            onRelease=function(t)
                local s = self.player.selected
                self.buttonPressed = true
                if t=="LOOT" then
                    s.target = obj
                    if s:inRange(1,obj) then
                        self:destroyPop()
                       -- s:attackTile(obj)
                    end
                elseif t=="PUSH" then
                    if s:inRange(1,obj) then
                        self:push(obj)
                    end
                end
            end
        })
        self.pop = pop
    end
end

function DungeonFloor:destroyPop()
    if self.pop then
        self.pop:destroy()
        self.pop = nil
    end
end

function DungeonFloor:push(obj,pusher)
    self:destroyPop()
    local psher = pusher or self.player.selected
    local xdiff,ydiff = obj:getX()-psher:getX(),obj:getY()-psher:getY()
    local p = obj:moveTo(obj:getX()+xdiff,obj:getY()+ydiff)
    log(p) log("!!??")
    return p
end
    
function DungeonFloor:rePopUp()
    self.camera:update(0.1)
    if not self.pop then
        return
    end
    self.pop:destroy()
    self:popUp(self._popUp[1],self._popUp[2])
end


function DungeonFloor:textinput(text)
  gooi.textinput(text)
end

function DungeonFloor:keypressed(key, scancode, isrepeat)
  gooi.keypressed(key, scancode, isrepeat)
  if key == "escape" then
    _LOG_FILE:close()
    love.event.quit()
  end
end

function DungeonFloor:keyreleased(key, scancode) log(key) log("kr")
  gooi.keyreleased(key, scancode)
end

return DungeonFloor