local floor = math.floor

GridMap = class:extend ("GridMap")

function GridMap:__init__(kwargs)
    kwargs = kwargs or {}
    self._kwargs = kwargs
    
    self._seed = kwargs.seed or os.time()
    math.randomseed(self._seed)
    self.seed = math.random()
    kwargs.parent = self
    self.paddy = Paddy:new(kwargs)
    self.usePaddy = kwargs.paddy or kwargs.usePaddy
    if self.usePaddy == nil then
        self.usePaddy = true
    end
    local h,w = 100,100
    self.joystick = Joystick:new({x=W()/2-w,y=H()-h,h=h,w=w})
    
    self.w = kwargs.w or 10
    self.h = kwargs. h or 10
    self.tileH = kwargs. tileH or 16
    self.tileW = kwargs.tiles or 16
    self.values = kwargs.values or {}
    self.observe = kwargs.observe or false
    
    self.turn = Toggle:new({"ENEMY","PLAYER"})
    self.turn.data = "PLAYER"
    self.turn.value = 0
    self.oldTurn = 0
    
    self.observations = {}  --turn:{tiles by id}
    self.map = {
        parent = self,
        height = self.h,
        width = self.w,
        draw = function(self)
            for x,tile in lume.iter(self.parent.allTiles) do
                tile:draw()
            end
        end,
        update = function (self, dt)
            for x,tile in lume.iter(self.parent.allTiles) do
            if tile.unit then assert(not tile.unit.isCamp) end
                tile:update()
            end
        end,
        convertPixelToTile = function(self,x,y)
            local xx,yy = x/self.parent.tileW,
                y/self.parent.tileH
            if xx<1 then xx = xx+1 end
            if yy<1 then yy = yy+1 end
            return xx,yy
            end,
        convertTileToPixel = function(self,x,y)
            return x*self.parent.tileW,
                y*self.parent.tileH
            end,
            tiles = {}
    }
    self.allTiles = {}
    self.grid = {}
    self.colorTiles = {}
    self.numGrid = {}
    self.mapGrid = {}
    self.camera = kwargs.camera or Camera()
    self.cameraMan = {x=0,y=0}
    
    self.player = {
        x=0,
        y=0,
        sight = 5,
        shaking = 0,
        animating = 0,
    }
    self.map.grid = {}
    self.map.tiles = {}
    self.grid = self.map.grid
    self.allTiles = {}
    self.fov = rot.FOV.Precise:new(function(fov,x,y) return self:lightCallback(fov,x,y) end, {topology=4})
    self.terrainFov = rot.FOV.Precise:new(function(fov,x,y) return self:tLightCallback(fov,x,y) end)
    self:loadTurn()
    
    --self:loadTiles()
end

function GridMap:load()
    self.fov = rot.FOV.Precise:new(function(fov,x,y) return self:lightCallback(fov,x,y) end, {topology=4})
    self.terrainFov = rot.FOV.Precise:new(function(fov,x,y) return self:tLightCallback(fov,x,y) end)
    
    for x,c in lume.iter(self.allTiles) do
        local x,y = c.x, c.y
        self.grid[x] = self.grid[x] or {}
        self.numGrid[x+1] = self.numGrid[x+1] or {}
        self.grid[x][y] = c
        self.mapGrid[y] = self.mapGrid[y] or {}
        self.mapGrid[y][x] = c
    end
    --[[for x in ume.iter self.grid) do
        for y,t in lume,iter(u) do
            self.mapGrid[y] = self.mapGrid[y] or {}
            self.mapGrid[y][x] = c
            -- log(x..", "..y)
            log("map at y: "..y..", x: "..x)
       end
    end]]
    
    self._grid = JumperGrid(self.mapGrid)
    self.pathfinder = PathFinder(self._grid,"ASTAR",
        function(val) return self:isWalkable(val) end
    )
    self.pathfinder:setMode("ORTHOGONAL")
end

function GridMap:addTile(...)
    self:loadTile(...)
end

function GridMap:shake(...)
    self.camera:setFollowStyle("LOCKON")
    self.camera:shake(...)
end

function GridMap:isDown(...)
    return self.paddy:isDown(...)
end

function GridMap:getTile(_tile,x,y,dontSave)
    local _tile = _tile or {}
    local u = nil
    local tile = (self.map.grid[x] and self.map.grid[x][y]) 
    if tile then return tile end
    tile = {
        x = x,
        y = y,
        w = self.tileW,
        h = self.tileH,
        isTile = true,
        getX = function(self)
            return self.x
        end,
        getY = function(self)
            return self.y
        end,
        getPos = function(self)
            return self.x,self.y
        end,
        properties = {
            collidable = false
        },
        map = self,
        id = _tile.id or x..", "..y,--lume.uuid(),
        _w, _h = 1,1,
        source = _tile.source,
        smell = 0,
        update = function(self,dt)
            if self.map.newTurn and self.smell>0 then
                tile.smell = tile.smell - .5
            elseif self.smell<0 then
                self.smell = 0
            end
            
            if self.unit ~= self.oldUnit then
                u = self.unit or self.oldUnit
                self.map:observeNoise(tile,1)
                self.map:observeSmell(tile,u.stats.smell)
                self.oldUnit = self.unit
            end
            if self.map.tmpselected == self then
                self.map:colorTile(self,"cyan")
            end
            self._w, self._h = resizeImage(game:getSource(self.source), self.w,self.h)
            self._x = self.x*self.w
            self._y = self.y*self.h
            if self.collider then
                self._x, self._y = self.collider:getPosition()
            end
        end,
        draw = function(self)
            love.graphics.draw(game:getSource(self.source),
                self._x,self._y,
                0,
                self._w, self._h
            )
            if self.color then
                local r, g, b, a = love.graphics.getColor()
                love.graphics.setColor(self.color)
                love.graphics.rectangle("fill",self._x,self._y,self.w,self.h)--,w,h)
                love.graphics.setColor(r,g,b,a)
            end
        end
    }
    if self.dontSave or dontSave then
        return tile
    end
    
    for _,x in pairs(tile) do
        _tile[_] = _tile[_] or x
        if type(x) == "function" then
            _tile[_] = x
        end
    end
    return _tile
end

function GridMap:loadTurn()
log(self.turn.value.."ftixx")
    self.observations[self.turn.value] = self.observations[self.turn.value] or {}
end

function GridMap:observeNoise(tile,volume)
    local obs = self.observations[self.turn.value]
    --warn(self.turn.value.."noise obs")
    obs[tile.id] = obs[tile.id] or {noise=0}
    obs[tile.id].noise = obs[tile.id].noise+volume
end

function GridMap:observeSmell(tile,smellLevel)
    --sorry, tile...
    tile.smell = smellLevel
end

function GridMap:loadTile(_tile,x,y)
    if not _tile then
        return
    end
    local tile = self:getTile(_tile,x,y)

    tile._x = tile.w*tile.x
    tile._y = tile.h*tile.y
    for u in pairs(_tile) do
        tile[u] = tile[u] or _tile[u]
    end
    tile.unit = nil
    
    self.map.grid[x] = self.map.grid[x] or {}
    self.map.grid[x][y] = tile
    self.map.tiles[tile.id] = tile
    tile.gid = tile.id
    table.insert(self.allTiles, tile)
    store(tile)
end

function GridMap:loadMap()
    self.map.grid = {}
    self.map.tiles = {}
    self.allTiles = {}
    for w = 1, self.w do
        self.map.grid[w] = self.map.grid[w] or {}
        for h = 1, self.h do
            local tile = self:getTile(nil,w,h)
            
            self.map.grid[w][h] = tile
            self.map.tiles[tile.id] = tile
            tile.gid = tile.id
            table.insert(self.allTiles, tile)
            local xx = (w+self.seed)/self.world.w
            local yy = (h+self.seed)/self.world.h
            local f = noise2D(xx,yy)
            
            for c,u in lume.iter(self.values) do
                local max, min, biome = u.max, u.min, u.biome
                if not max then
                    max, min, biome = u[1], u[2], u[3]
                end
                if f <= max and f  >min then
                    tile.biome = lume.copy(biome)
                    tile.source = biome.source
                    break
                end
            end
            
        end
    end
    
end
                    
            
            
function GridMap:isWalkable(val)
    if val.collidable then
        return false
    else
        return true
    end
end

--returns false if light can't pass through tile
function GridMap:lightCallback(fov, x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    if tile.unit then return false end
    return true
end
function GridMap:tLightCallback(fov, x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    --if tile.unit then return false end
    return true
end

function GridMap:onFOVCompute(x, y, range, visibility)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    --self:makeTileDarker(tile)
    
end
        
function GridMap:findPathTo(obj,x,y)

    local obj = obj
    local x = x
    local y = y
    if type(obj) ~= "table" then
        x = obj
        y = x
        obj = y
    end
    
    local floor = math.floor
    local _x,_y = obj:getPos()
    start = {x=floor(_x),y=floor(_y)}
    
    if not self.grid[floor(x)] then log("no tile,returning") return end
    if not self.grid[floor(x)][floor(y)] then log("no tile, returning") return end
    if self.grid[floor(x)][floor(y)].collidable == true then
        log("collidable")
        return
    end
    obj._goto = self.grid[floor(x)][floor(y)]
    self.c_unit = obj
    local path = self.pathfinder:getPath(start.x,start.y,floor(x),floor(y))
    if not path then
        sprint = x..", "..y..": "..tostring(self.grid[floor(x)][floor(y)].unit)
    end

    _path = {}
    if path then for x,c in path:nodes() do log(c) table.insert(_path,x) end end
    if not path then cE = (cE or 0)+1 end
    --if cE and cE > 15 then log("cE returning") return false end-- error("path.tsktsk") end
    local f = " at "..tostring(obj)
    if path then log("path is "..#path._nodes..f) else log("path is not "..f) end
    return path and path._nodes
end

function GridMap:draw()
    local trash = self.unitDisplay and self.unitDisplay:draw()
    self.camera:attach()
    self.map:draw()
    self.camera:detach()
    self.camera:attach()
    for nx,i in pairs(self.colorTiles) do
        self:_colorTile(game:get(nx),i)
    end
    self.camera:detach()
    local trash = self.usePaddy and self.paddy:draw()
end

function GridMap:update(dt)
    local trash = self.unitDisplay and self.unitDisplay:update(dt)
    if self.turn.value ~= self.oldTurn then
        self.oldTurn = self.turn.value
        self.newTurn = true
        self:loadTurn()
    else
        self.newTurn = false
    end
    
    local trash = self.usePaddy and self.paddy:update(dt)
    local gur = function(...) self:onFOVCompute(...) end
    self.fov:compute(self.player.x,self.player.y,self.player.sight,
        gur)
    if not self.camera.shaking then
        self.camera:setFollowStyle("none")
    end
    self.camera:follow(self.cameraMan.x,self.cameraMan.y)
    self.camera:update(dt)
    self.map:update(dt)
     
    if self.cameraMan.animation then
        self.cameraMan.animation:update(dt)
        if self.cameraMan.animation.done then
            self.cameraMan.animation = nil
        end
    end
end

function GridMap:getTileWithPos(x,y)
    if type(x) == "table" then
        local x = x.x or x[1]
        local y = x.y or x[2]
    end
    local x, y = self.map:convertPixelToTile(x,y)
    local x, y = floor(x), floor(y)
    log(x..","..y)
    return self.grid[x] and self.grid[x][y]
end

function GridMap:old_colorTile(tile,color)
    if type(color) == "string" then
        color = game:getColor(color)
    end
    tile.color = color
end

function GridMap:old_uncolorTile(tile)
    tile.color = nil
end

function GridMap:colorTile(tile,color)
    if tile.tile then error()
        tile = tile.tile
    end
    tile.colored = true
    if tile.y==7 then ggh=the end
    --table.insert(self.colorTiles,{tile,color})
    
    self.colorTiles[tostring(tile)] = color
end

function GridMap:uncolorTile(tile,color)
    if not self.colorTiles then
       -- return
    end
    self.colorTiles[tostring(tile)] = nil
    --[[
    local tmC = {}
    for x,i in ipairs(self.colorTiles) do
        if i[1] ~= tile and ((color and color==i[2]) or not color) then
            table.insert(tmC, i)
            
        end
    end
    self.colorTiles = tmC]]
end

function GridMap:_colorTile(tile,color)
    if not tile then
        return
    end
    
    if type(color) == "string" then
        color = game:getColor(color)
    end
    
    local x, y = tile._x, tile._y
    local _tile = tile--self.map.tiles[tile.gid]
    local w, h = _tile.width, _tile.height 
    local r, g, b, a = love.graphics.getColor()
    love.graphics.setColor(color)
    love.graphics.rectangle("fill",x,y,16,16)--,w,h)
    love.graphics.setColor(r,g,b,a)
end

function GridMap:makeTileDarker(tile)
    self:colorTile(tile,{0.5,0.5,0.5,1})
end

function GridMap:colorAllTiles(color)
    self.colorTiles = {}
    
    grid = lume.copy(self.grid)
    table.insert(grid,self.grid[0])
    for x,jc in ipairs(grid) do 
        _jc = lume.copy(jc)
        table.insert(_jc,grid[x][0])
        for u,c in ipairs(_jc) do
            self:colorTile(c,color)
        end
    end
end


function GridMap:mousepressed(...)
    --local trash = self.usePaddy and self.paddy:mousepressed(...)
    local trash = self.unitDisplay and self.unitDisplay:mousepressed(...)
end

function GridMap:mousereleased(...)
    --local trash = self.usePaddy and self.paddy:released(...)
    local trash = self.unitDisplay and self.unitDisplay:mousereleased(...)
end

function GridMap:touchpressed(...)
    local trash = self.usePaddy and self.paddy:touchpressed(...)
    local trash = self.unitDisplay and self.unitDisplay:touchpressed(...)
end

function GridMap:touchreleased(...)
    local trash = self.usePaddy and self.paddy:touchreleased(...)
    local trash = self.unitDisplay and self.unitDisplay:touchreleased(...)
end

function GridMap:touchmoved(...)
    local trash = self.usePaddy and self.paddy:touchmoved(...)
end


return GridMap