sti = require "tools/sti"
rot = require "tools/src.rot"
JumperGrid = require "tools/jumper.grid"
PathFinder = require "tools/jumper.pathfinder"


local floor = math.floor
MapObject = class:extend("MapObject")

function MapObject:__init__(source,plugins)
    self.map = sti(source,plugins)
    self.grid = {}
    self.colorTiles = {}
    --[[self.grid.__ipairs = function(i)
    return grid[0]]--
    self.numGrid = {}
    self.mapGrid = {}
    self.camera = Camera()
    self.cameraMan = {x=0,y=0}
    self.player = {
        x=0,
        y=0,
        sight = 5
    }
    self.fov = rot.FOV.Precise:new(function(fov,x,y) return self:lightCallback(fov,x,y) end, {topology=4})
    self.terrainFov = rot.FOV.Precise:new(function(fov,x,y) return self:tLightCallback(fov,x,y) end)
    
    for x,c in lume.iter(self.map.allTileInstances) do
        local x,y = self.map:convertPixelToTile(c.x,c.y)
        local x = floor(x)
        local y = floor(y)
        self.grid[x] = self.grid[x] or {}
        self.numGrid[x+1] = self.numGrid[x+1] or {}
        self.grid[x][y] = c
        self.mapGrid[y] = self.mapGrid[y] or {}
        self.mapGrid[y][x] = c
       -- log(x..", "..y)
    end
    self._grid = JumperGrid(self.mapGrid)
    self.pathfinder = PathFinder(self._grid,"ASTAR",
        function(val) return self:isWalkable(val) end
    )
    self.pathfinder:setMode("ORTHOGONAL")
    log({self.grid[0][0].x/16,self.grid[0][0].y/16})
    assert(self._grid._nodes[9])--,lume.serialize(self._grid))
end

function MapObject:isWalkable(val)
    if self.map.tiles[val.gid].properties.collidable then
        return false
    else
        return true
    end
end

--returns false if light can't pass through tile
function MapObject:lightCallback(fov, x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    if tile.unit then return false end
    return true
end
function MapObject:tLightCallback(fov, x,y)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    if tile.unit then return false end
    return true
end

function MapObject:onFOVCompute(x, y, range, visibility)
    local tile = self.grid[x] and self.grid[x][y]
    if not tile then return end
    --self:makeTileDarker(tile)
    
end
        
function MapObject:findPathTo(obj,x,y)

    local obj = obj
    local x = x
    local y = y
    if type(obj) ~= "table" then
        x = obj
        y = x
        obj = y
    end
    x=x
    y=y
    local floor = math.floor
    local _x,_y = self.map:convertPixelToTile(obj.tile.x,obj.tile.y)
    sprint = x..", "..y
    _x = _x
    _y = _y
    start = {x=floor(_x),y=floor(_y)}
    
    if not self.grid[floor(x)] then return end
    if not self.grid[floor(x)][floor(y)] then return end
    if self.map.tiles[self.grid[floor(x)][floor(y)].gid].properties.collidable == true then
        return
    end
    self._c_unit = obj
    local path = self.pathfinder:getPath(start.x,start.y,floor(x),floor(y))
    if not path then
        sprint = x..", "..y..": "..tostring(self.grid[floor(x)][floor(y)].unit)
    end
    _path = {}
    if path then for x,c in path:nodes() do table.insert(_path,x) end end
    log("......"..#_path)
    if not path then cE = (cE or 0)+1 end
    if cE and cE > 15 then return false end-- error("path.tsktsk") end
    return path and path._nodes
end

function MapObject:draw()
    self.camera:attach()
    self.map:draw()
    self.camera:detach()
    self.camera:attach()
    for nx,i in lume.iter(self.colorTiles) do
        self:_colorTile(i[1],i[2])
    end
    self.camera:detach()
end

function MapObject:update(dt)
    self.fov:compute(self.player.x,self.player.y,self.player.sight,
        function(...) self:onFOVCompute(...) end)
    self.camera:follow(self.cameraMan.x,self.cameraMan.y)
    self.camera:update(dt)
    self.map:update(dt)
end

function MapObject:getTileWithPos(x,y)
    if type(x) == "table" then
        local x = x.x or x[1]
        local y = x.y or x[2]
    end
    local x, y = self.map:convertPixelToTile(x,y)
    local x, y = floor(x), floor(y)
    log(x..","..y)
    return self.grid[x] and self.grid[x][y]
end

function MapObject:colorTile(tile,color)
    table.insert(self.colorTiles,{tile,color})
end

function MapObject:_colorTile(tile,color)
    if not tile then
        return
    end
    
    if type(color) == "string" then
        color = game:getColor(color)
    end
    
    local x, y = tile.x, tile.y
    local _tile = self.map.tiles[tile.gid]
    local w, h = _tile.width, _tile.height 
    local r, g, b, a = love.graphics.getColor()
    love.graphics.setColor(color)
    love.graphics.rectangle("fill",x,y,16,16)--,w,h)
    love.graphics.setColor(r,g,b,a)
end

function MapObject:makeTileDarker(tile)
    self:colorTile(tile,{0.5,0.5,0.5,1})
end

function MapObject:colorAllTiles(color)
    self.colorTiles = {}
    
    grid = lume.copy(self.grid)
    table.insert(grid,self.grid[0])
    for x,jc in ipairs(grid) do 
        _jc = lume.copy(jc)
        table.insert(_jc,grid[x][0])
        for u,c in ipairs(_jc) do
            self:colorTile(c,color)
        end
    end
end

return MapObject