local Underground = GridMap:extend("Underground")
local Entity = class:extend("Entitiy")

function Entity:__init__(map,kwargs)
    self.active = true
    
    kwargs = kwargs or {}
    self._kwargs = kwargs
    self.angle = kwargs.angle or 0
    self._w,self._h = 0,0
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    self.h = kwargs.h or 16
    self.w = kwargs.w or 16
    self.ax = kwargs.ax or 0
    self.ay = kwargs.ay or 0
    
    self.sight = kwargs.sight or 10
    self.collisionType = kwargs.collisionType or kwargs.colType or "touch"
    
    self.source = kwargs.source or "player.png"
    
    self.map = map
    self.world = self.map.world
    self.world:add(self,self.x,self.y,self.w,self.h)

    self.speed = {x=0,y=0}
    self.weight = 400
    self.gravity = 0
    self.initGravity = 300

    self._speed = 100
end

function Entity.checkCollision(item, other)
    if not other.isBackground then
        return item.collisionType
    end
end

function Entity:move(x,y)
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    
    return actualX, actualY
end

function Entity:updateGravity(dt)
    
    self.gravity = self.gravity+self.weight*dt
    self.speed.y = self.speed.y + self.gravity*dt
end

function Entity:update(dt)
    self.speed = {x=0,y=0}
    self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
    
    self:updateGravity(dt)
    
    local vx, vy = self.x+self.speed.x*dt,self.y+self.speed.y
    
    if 1 then --???
        local mx, my = self:move(vx,vy)
        self.x, self.y = mx,my
    end
end

function Entity:draw()
    self.map.camera:attach()
    love.graphics.draw(game:getSource(self.source),self.x,self.y,self.angle,
    self._w,self._h,self.w/2, self.h/2)
    self.map.camera:detach()
end

function Entity:dir(n)
    if n > 0 then return 1 elseif n < 0 then return -1
    else return 0 end
end

function Entity:destroy()
    self.world:remove(self)
    self.source = "none.png"
    self.active = false
end

function Entity:accelerate(x,y)
    self.ax = self.ax + x
    self.ay = self.ay + y
end

local Bullet = Entity:extend("Bullet")

function Bullet:__init__(map,kwargs)
    self.isBullet = true
    
    self.super.__init__(self,map,kwargs)
    self.source = "bullet.png"
    self.parent = kwargs.parent or error("no parent")
    assert(self.angle>0)
    self.angleY = math.sin(self.angle)
    self.angleX = math.cos(self.angle)
    self.ac = 500
    
end

function Bullet:update(dt)
    self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
   -- self._h=self._h*-1
   -- self._w=self._w*-1
    local s = self.speed.y
    self.speed.y = 0
    self:updateGravity(dt)
    
    self.speed.y=0
    self.speed.x = 100+self.ax
    self.speed.y = -100+self.ay
    self:accelerate(self.ac*dt,self.ac*dt*-1)
    
    local vx, vy = self.x+self.speed.x*dt*self.angleX,
        self.y+self.speed.y*dt*self.angleY
    
    if 1 then --???
        local mx, my = self:move(vx,vy)
        self.x, self.y = mx,my
    end
end

function Bullet.checkCollision(item, other)
    if other.isTile and not other.isBackground then
        other.life = other.life - 1
        item.die = true
        return item.collisionType
    end
    
    if other == item.parent then
        return nil
    elseif other.isBullet then
        return nil
    elseif not other.isBackground then
        return item.collisionType
    end


end

function Bullet:move(x,y)
    self.die = false
    if not self.active then
        return 0,0
    end
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    if self.die then
        self:destroy()
    end
    
    return actualX, actualY
end

function Bullet:draw()
    self.map.camera:attach()
    
    local v = game:getSource(self.source)
    local xx,yy = v:getWidth()/2, v:getHeight()/2
    local ca = math.pi^2
    --local ang = math.deg(self.angle)
    
    love.graphics.draw(game:getSource(self.source),self.x+xx,self.y+yy,math.pi-(self.angle+ca),
    self._w,self._h,xx,yy)
    self.map.camera:detach()
    --love.graphics.print(self.angleX..", "..self.angleY,self.x,self.y)
    
end
local myPlayer = Entity:extend("myPlayer")

function myPlayer:__init__(map,kwargs)
    kwargs = kwargs or {}
    self._kwargs = kwargs
    self.angle = 0---2--1+math.pi/2--kwargs.angle or 0
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    self.h = kwargs.h or 16
    self.w = kwargs.w or 16
    self.sight = kwargs.sight or 10
    self.source = kwargs.source or "player.png"
    self.paddy = Paddy:new({joysticks=true})
    
    self.Ljoystick = self.paddy.Ljoystick
    self.Rjoystick = self.paddy.Rjoystick
    
    self.map = map
    self.world = self.map.world
    self.world:add(self,self.x,self.y,self.w,self.h)
    
    --self.collider = self.map.world:newRectangleCollider(self.x,self.y,self.w,self.h)

    self.speed = {x=0,y=0}
    self.weight = 400
    self.gravity = 0
    self.initGravity = 300

    self._speed = 100
    self.bullets = {}
    self.bulletLen = 0
end

function myPlayer:shoot(angle)
    local newBullet = Bullet:new(self.map,{
        x = self.x,
        y = self.y,--+self.h,
        weight = 10,
        w = 10,
        h = 5,
        parent = self,
        angle = angle+math.pi*0
    }) --warn(angle)
    self.bullets[self.bulletLen+1] = newBullet
    self.bulletLen = self.bulletLen+1
end

function myPlayer.checkCollision(item, other)
    if other.isBackground then
        return nil
    elseif other.isBullet and other.parent == item then
        return
    else
        return "slide"
    end
end

function myPlayer:move(x,y)
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    
    local col
    local ignore = false
    for i = 1, len do
        col = collisions[i]
        if col.normal.y == -1 or col.normal.x ~=0 then
            self.canJump = true
            if col.normal.y == -1 and self.gravity > 0 then
                ignore = true
                self.gravity = 0
            end
            self.jumping = falsed
        end
        if col.normal.y == 1 then
            self.gravity = self.initGravity/2
            self.canJump = false
        end
    end
    if self.gravity>self.initGravity*.45 and not ignore then
        self.canJump = false
    end
    
    return actualX, actualY
end

function myPlayer:jump()
    if self.canJump then
        self.jumping = true
        self.gravity = -self.initGravity
    end
end

function myPlayer:update(dt)
    self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
    --self.x, self.y = self.collider:getPosition()
    --self.anglxe = self.collider:getAngle() --?????change back
    
    for i = 1, self.bulletLen do
        local b = self.bullets[i]
        b:update(dt)
    end
    
    self.paddy:update(dt)
    
    self.speed = {x=0,y=0}
    
    self.gravity = self.gravity+self.weight*dt
    
    self.speed.y = self.gravity*dt

    self.speed.x = self.Ljoystick:getXDir()*self._speed
  --  self.speed.y = self.speed.y + self.Ljoystick:getYDir()*self._speed
    if self.Ljoystick:getYDir()*-1 > 0 and self.canJump then
        self.gravity = -300
    end
    
    local angle = self.Rjoystick.angle
    if angle ~= 0 and not sd then
        self:shoot(angle)
    end
    
    if self.paddy:isDown("left") then
        self.speed.x = -self._speed
    elseif self.paddy:isDown("right") then
        self.speed.x = self._speed
    elseif self.paddy:isDown("up") then
        self.gravity = -300 --self.speed.y = self.speed.y-self._speed*3
    end
    
    local x,y = 0,0--(0self.paddy.joystick:getX(), self.paddy.joystick:getY()
    local vx, vy = self.x+self.speed.x*dt,self.y+self.speed.y
    
    self.canJump = false
    if 3 then--self.speed.y~=0 or self.speed.x~=0 then
        local mx, my = self:move(vx,vy)
        self.x, self.y = mx,my
    end
  --  self.collider:applyLinearImpulse(x,y)
    self.map.camera.scale=1
end

function myPlayer:draw()

    for i = 1, self.bulletLen do
        local b = self.bullets[i]
        b:draw()
    end

    self.map.camera:attach()
    
    local v = game:getSource(self.source)
    local xx,yy = v:getWidth()/2, v:getHeight()/2
    
    love.graphics.draw(game:getSource(self.source),self.x+xx,self.y+yy,self.angle,
    self._w,self._h,xx,yy)
    self.map.camera:detach()
    self.paddy:draw()
end

function myPlayer:released(...)
    self.paddy:touchreleased(...)
end

function myPlayer:pressed(...)
    self.paddy:touchpressed(...)
end

function myPlayer:touchmoved(...)
    self.paddy:touchmoved(...)
end

function Underground:log(l)
    log(l)
end


function Underground:enter()
 --   self:__init__()
end

--[[
    y
x       b
    a
]]
function Underground:__init__(kwargs)
    local kwargs = kwargs or {}
    self.super.__init__(self,kwargs)
    self.cameraSpeed = 0.02
    self.usePaddy = false
    self.world = bump.newWorld(16)
    --self.world:addCollisionClass("Dirt", {ignores={"Dirt"}})
    --self.world:addCollisionClass("Player")
    self.camera.scale=.5
    self:loadMap(self.w, self.h)
    self:spawnPlayer()
end

function Underground:loadMap(w,h)
    for x = 1,w do
        for y = 1,h do
            local tile = self:getTile({},x,y)
            tile.source = "dirt.png"
            local tx, ty = x*tile.w, y*tile.h
            --tile.collider = self.world:newRectangleCollider(tx,ty,tile.w,tile.h)
            tile.collider = {
                getPosition = function()
                    return self.world:getRect(tile)
                end,
                destroy = function() end
            }
            
            --tile.collider:setObject(tile)
            --tile.collider:setType("static")
            --tile.collider:setCollisionClass("Dirt")
            
            self:addTile(tile,x,y)
            tile = self:getTile(nil,x,y)
            self.world:add(tile,tx,ty,tile.w,tile.h)

        end
    end
end

--removes a tile, not asvised to use
function Underground:removeTile(tile,y)
    if type(tile) == "number" then
        tile = self:getTile(tile,y)
    end
    if tile.collider then
        tile.collider:destroy()
    end
    self.world:remove(tile)
    self.grid[tile.x][tile.y] = nil
end

function Underground:spawnPlayer()
    local x = math.random(1+5,self.w-5)
    local y = math.random(1+5,self.h-5)
    self.player = myPlayer:new(self,{x=x*16,y=y*16})
    local tile = self:getTile(nil,x,y)
    tile.source="none.png"
    tile.isBackground = true
    --tile.collider = nil
    for r,o in ipairs(getRound(5)) do
        local t = self:getTile(nil,x+o.x,y+o.y)
        if t.collider then t.collider:destroy() t.collider=nil end
        t.isBackground = true
        t.source = "none.png"
        
    end
end

function Underground:update(dt)
    self:callSuper("update",dt)
    self.player:update(dt)
    self.cameraMan = {x=self.player.x,y=self.player.y}
  --  self.world:update(dt)
    
    local maxZoom = 10
    local minZoom = 0.05
    if self.camera.scale >  maxZoom then
        self.camera.scale = maxZoom-1
        self.cameraMan.animation= nil
    elseif self.camera.scale < minZoom then
        self.camera.scale = minZoom+0.01
    end
        
end

function Underground:draw()
    self:callSuper("draw")
    self.camera:attach()
    self.camera:detach()
    self.player:draw()
    
    self.camera:draw()
end

function Underground:getTile(_tile,x,y,dontSave)
    local _tile = _tile or {}
    local u = nil
    local tile = (self.map.grid[x] and self.map.grid[x][y]) 
    if tile then return tile end
    tile = {
        active = true,
        life = 5,
        x = x,
        y = y,
        w = self.tileW,
        h = self.tileH,
        isTile = true,
        getX = function(self)
            return self.x
        end,
        getY = function(self)
            return self.y
        end,
        getPos = function(self)
            return self.x,self.y
        end,
        properties = {
            collidable = false
        },
        map = self,
        id = _tile.id or x..", "..y,--lume.uuid(),
        _w, _h = 1,1,
        source = _tile.source,
        smell = 0,
        destroy = function(self)
            
            self.map.world:remove(self)
            self._x,self._y=0,0
            self.active = false
            self.source = "none.png"
            self.collider = nil
        end,
        update = function(self,dt)
            if self.life < 0 and self.active then
                self:destroy()
            end
            
            if self.map.newTurn and self.smell>0 then
                tile.smell = tile.smell - .5
            elseif self.smell<0 then
                self.smell = 0
            end
            
            if self.unit ~= self.oldUnit then
                u = self.unit or self.oldUnit
                self.map:observeNoise(tile,1)
                self.map:observeSmell(tile,u.stats.smell)
                self.oldUnit = self.unit
            end
            if self.map.tmpselected == self then
                self.map:colorTile(self,"cyan")
            end
            self._w, self._h = resizeImage(game:getSource(self.source), self.w,self.h)
            self._x = self.x*self.w
            self._y = self.y*self.h
            if self.collider then
                self._x, self._y = self.collider:getPosition()
            end
        end,
        draw = function(self)
            love.graphics.draw(game:getSource(self.source),
                self._x,self._y,
                0,
                self._w, self._h
            )
            if self.color then
                local r, g, b, a = love.graphics.getColor()
                love.graphics.setColor(self.color)
                love.graphics.rectangle("fill",self._x,self._y,self.w,self.h)--,w,h)
                love.graphics.setColor(r,g,b,a)
            end
        end
    }
    if self.dontSave or dontSave then
        return tile
    end
    
    for _,x in pairs(tile) do
        _tile[_] = _tile[_] or x
        if type(x) == "function" then
            _tile[_] = x
        end
    end
    return _tile
end

function Underground:mousereleased(x, y, button)
    local trash = self.pop and self.pop:mousereleased(x,y,button)
    self.player:released()
    gooi.released()
end

function Underground:mousepressed(x, y, button)
    local trash = self.pop and self.pop:mousepressed(x,y,button)
   -- self.player:pressed()
    gooi.pressed()
end

function Underground:touchpressed(id, x, y, dx, dy, pressure)
    self.super.touchpressed(self,id,x,y,dx,dy,pressure)
    
    self.player:pressed(id,x,y,dx,dy,pressure)
end

function Underground:touchmoved(...)
    self.player:touchmoved(...)
end

function Underground:touchreleased(id, x, y, dx, dy, pressure)
    self.super.touchreleased(self,id,x,y,dx,dy,pressure)
    
    self.player:released(id,x,y,dx,dy,pressure)
    
    local trash = self.pop and self.pop:touchreleased(id,x,y,dx,dy,pressure)
    if trash then
        self.buttonPressed = true
    end
    if not self.buttonPressed then
        ttx = x
        tty = y
    else
        self.buttonPressed = false
        return
    end
    x,y = self.camera:toWorldCoords(x,y)
   
    local tile = self:getTileWithPos(x,y)
    local sx,sy = self.map:convertPixelToTile(x,y)
    sx = math.floor(sx)
    sy = math.floor(sy)
end

function Underground:textinput(text)
  gooi.textinput(text)
end

function Underground:keypressed(key, scancode, isrepeat)
  gooi.keypressed(key, scancode, isrepeat)
  if key == "escape" then
    p:close()
    love.event.quit()
  end
end

function Underground:keyreleased(key, scancode) log(key) log("kr")
  gooi.keyreleased(key, scancode)
end

return Underground