local UnitDisplay = class:extend("UnitDisplay")
local unknown = "???"

--[[Grid:
Unit's image|name
            |species
            |job
            |alias
____________|...

]]

function UnitDisplay:__init__(kwargs)
    local kwargs = kwargs or {}
    
    self.unit = kwargs.unit or kwargs
    
    self.health = kwargs.health or unknown
    self.strength = kwargs.strength or unknown
    self.agility = kwargs.agility or unknown
    self.target = kwargs.target or "none"
    self.favAttack = kwargs.favAttack or unknown
    self.favTeammate = kwargs.favTeammate or unknown
    self.name = kwargs.name or unknown
    self.species = kwargs.species or unknown
    self.alias = kwargs.alias or unknown
    self.job = kwargs.job or unknown
    
    self.source = kwargs.source or self.unit.source or "none.png"
    self.img = type(self.source) ~= "userdata" and game:getSource(self.source) or self.source
    
    
    local len = kwargs.data and #kwargs.data
    self.listbox = lume.copy(listbox)
    self.gooi = kwargs.gooi
    self.h = kwargs.h or H()
    self.w = kwargs.w or W()/6
    
    self.order = {
        "health", "strength", "agility", "target", "ATTACKS",
        "FAVORITES", "attack", "teammate"
    }
    
    self.data = {
        health = self.health,
        strength = self.strength,
        agility = self.agility,
        target = self.target,
        ATTACKS = "",
        FAVORITES = "",
        attack = self.favAttack,
        teammate = self.favTeammate
    }
    
    self.listbox:init()
    
    self.otherButton = gooi.newButton({text="..."}):
    onRelease(function(b)
        local t = b.text
    end)
    self.placeLabel = gooi.newLabel({text=""})
    
    self.infoPanel = gooi.newPanel({
        w=self.w, h=self.h/3,
        x=kwargs.x or 0,  y=kwargs.y or 0,
      --  group="displayPanel",
        layout="grid 5x2"
    }):setRowspan(1,1,5)
    :add(
        self.placeLabel,
        gooi.newLabel({text=self.name}),
        gooi.newLabel({text=self.species}),
        gooi.newLabel({text=self.job}),
        gooi.newLabel({text=self.alias}),
        self.otherButton
    )

    self._data = {
        x = kwargs.x or 0,
        y = (kwargs.y or 0)+self.h/3,
        w = self.w,
        h = self.h - (self.h/6),
        istouch = true,
        visible = true,
        incenter = false,
        enabled = true,
        selectedcolor = game:getColor("red"),
        bgcolor = game:getColor("black"),
        font =love.graphics.newFont(29)
    }
    
    --self._data.h = 16--self._data.h - self._data.h/(len+1)
    --self._data.h = self._data.h/2

    self.imgW, self.imgH = resizeImage(self.img,self.placeLabel.w, self.placeLabel.h)
    
    self.listbox:newprop(self._data)
    self.add = function(...) self.listbox:additem(...) end
    
    for _,x in ipairs(self.order) do
        self.add(x..": "..self.data[x],_)
    end
    
    self.listbox:setselected(1)
    
    self:create()

end

function UnitDisplay:destroy()
    self.listbox:setenabled(false)
    self.listbox:setvisible(false)
    local gooi = self.gooi or gooi
    gooi.removeComponent(self.infoPanel)
end

function UnitDisplay:create()
    gooi.removeComponent(self.infoPanel)
    self.listbox:setenabled(true)
    self.listbox:setvisible(true)
    gooi.addComponent(self.infoPanel)
end

function UnitDisplay:getSelectedText()
    return self.listbox:gettext(self.listbox:getselected())
end

function UnitDisplay:getSelectedDataText()
    return self.listbox:getdata(self.listbox:getselected())
end

function UnitDisplay:draw()
    self.listbox:draw()
    
    gooi.draw()
  --  error(self.placeLabel.x..", "..self.placeLabel.y..":w,h: "..self.imgW..", "..self.imgH)
    love.graphics.draw(self.img,self.placeLabel.x, self.placeLabel.y, 0, self.imgW, self.imgH)
    local trash = self.gooi and self.gooi.draw()
end

function UnitDisplay:update(dt)

    self.listbox:update(dt)
    local trash = self.gooi and self.gooi.update(dt)
    self.otherButton:setText(self:getSelectedText())

end

function UnitDisplay:mousepressed(x,y,button)
    local trash = self.gooi and self.gooi.pressed()
end

function UnitDisplay:mousereleased(x,y,button)
    local trash = self.gooi and self.gooi.released()
end

function UnitDisplay:touchpressed(id, x, y, dx, dy, pressure)

    self.listbox:press(x,y)
    if (x>self.listbox.x and x<self.listbox.x+self.listbox.w) and (y>self.listbox.y and y<self.listbox.y+self.listbox.h) then
        return true
    end
    

end

function UnitDisplay:touchreleased(id, x, y, dx, dy, pressure)

    self.listbox:released(x,y)
    if (x>self.listbox.x and x<self.listbox.x+self.listbox.w) and (y>self.listbox.y and y<self.listbox.y+self.listbox.h) then
        return true
    end

end

return UnitDisplay