function scale(_noise)
    -- Rescale from -1.0:+1.0 to 0.0:1.0
    return _noise--/2.0+0.5
end


function getNoise(x,y,options)

    local options = {
        octaves = (options and options.octave) or 1,
        frequency = (options and options.frequency) or 1,
        amplitude = (options and options.amplitude) or 1,
        persistence = (options and options.persistence) or 1,
        power = (options and options.power) or 1
    }
    
    local n = 0
    local fr = options.frequency
    local oct = options.octaves
    local a = options.amplitude
    local p = options.persistence
    local pow = options.power
    
    for i = 1,oct do
        n = n+noise(x*fr,y*fr)*a
        log(n)
        fr = fr*2
        a = a*p
    end
    local n = math.pow(scale(n,{max=oct}),pow)
    return n
end

function loadMap()
    map = {}
    h = mapSize.value
    w = mapSize.value
    local sx, sy
    for x=1,60 do
        for y=1,60 do
            sx = x/w
            sy = y/h
            values = {
                octave=octaves.value,
                frequency=frequency.value+(miniFrequency.value/100),
                amplitude=amplitude.value+(miniAmplitude.value/100),
                persistence=persistence.value+(miniPersistence.value/100),
                power=pow.value
            }
            local t = {x=x,y=y,noise=getNoise(sx,sy,values)}
            table.insert(map,t)
        end
    end
    return map
end

function love.load()
    math.randomseed(5)
    class = require 'tools/class'

    require 'tools/gooi'
    require 'tools/tools'

    lume = require 'tools/lume'

    --biome sources
    Biome = require 'noise/biomes'
    
    --params for noise
    mapGen = require "noise/mapGen"
    Game=require "game"
    game=Game:new()
    game:__init__()
    map = {}
    
    mapSize = gooi.newSpinner({
        value=10,
        min=0,
        max=1000
    })
    octaves = gooi.newSpinner({
        value=1,
        min=0,
        max=100
    })
    frequency = gooi.newSpinner({
        value=1,
        min=0,
        max=100
    })
    miniFrequency = gooi.newSpinner({
        value=0,
        min=0,
        max=100
    })
    amplitude = gooi.newSpinner({
        value=1,
        min=0,
        max=100
    })
    miniAmplitude = gooi.newSpinner({
        value=0,
        min=0,
        max=100
    })
    persistence = gooi.newSpinner({
        value=1,
        min=0,
        max=100
    })
    miniPersistence = gooi.newSpinner({
        value=0,
        min=0,
        max=100
    })
    pow = gooi.newSpinner({
        value=1,
        min=0,
        max=1000
    })
    editor = gooi.newPanel({x=W()-(W()/5),y=0,w=W()/5,h=H(),layout="grid 20x1"}):
    add(
        gooi.newButton({text="mapSize:"}),
        mapSize,
        gooi.newButton({text="octaves:"}),
        octaves,
        gooi.newButton({text="frequency:"}),
        frequency,
        gooi.newButton({text=".frequency:"}),
        miniFrequency,
        gooi.newButton({text="amplitude:"}),
        amplitude,
        gooi.newButton({text=".amplitude:"}),
        miniAmplitude,
        gooi.newButton({text="persistence:"}),
        persistence,
        gooi.newButton({text=".persistence:"}),
        miniPersistence,
        gooi.newButton({text="power:"}),
        pow,--[[gooi.newRadio({text="math.pow?",radioGrouo="noise"}):
             onSelected(function(y)
                if y.selected then
                    pow = true
                else
                    pow = false
                end
            end)]]
        gooi.newButton({text="getHeight"}):
            onRelease(function(y)
                reload(_G[y.text])
                _func = _G[y.text]
            end),
        gooi.newButton({text="getMoisture"}):
            onRelease(function(y)
                _func=_G[y.text]            
                reload(_G[y.text])
            end)
    )
    _func = getHeight
    setters = {octaves,frequency,amplitude,
    miniFrequency,miniAmplitude,persistence,miniPersistence,
    mapSize,pow}
    
end

function love.update(dt)
    gooi.update(dt)
    d=0
    local l = false
    for x,i in ipairs(setters) do
        if i.value ~= i.oldVal then
        
            i.oldVal = i.value
            l = true
            ii=i
        end
    end

    if l then 
        loadMap()
    else
    end
            
end

function love.draw()
    gooi.draw()
    for i,x in ipairs(map) do
        love.graphics.setColor({255,255,255,1*x.noise})--*x.noise
        love.graphics.rectangle("fill",x.x*16,x.y*16,16,16)
    end
end

function love.mousepressed()
    gooi.pressed()
end

function love.mousereleased()
    gooi.released()
end