--Animation lua module ported from kivy.
local None = nil
local True = true
local False = false
Clock = class:extend("Clock")
Clock.updates = {}
function Clock.schedule_interval(item,func,interval)
    --Who cares bout the interval for now
    table.insert(Clock.updates,{func,item})
    return _Checker:new(func)
end
function Clock.update(dt)
    for _,func in ipairs(Clock.updates) do
        func[1](func[2],dt)
    end
end

_Checker = class:extend("Checker")
function _Checker:__init__(func)
    self.func = func
end
function _Checker:cancel()
    lume.remove(Clock.updates,self.func)
    self = nil --?
end

local onComplete = "onComplete"
local onProgress = "onProgress"
local onStart = "onStart"
Animation = class:extend("Animation")
list = class:extend("list")
function list:add(item)
    table.insert(self,item)
end

function list:remove(item)
    lume:remove(self,item)
end

function list:hasItem(item)
    return lume.find(self,item)
end

function list:__init__()
end
Animation._instances = list:new()
Animation.help = {}
--[[Animation.help.main = '\
Animation
=========

 class `Animation` and  class `AnimationTransition` are used to animate
 class `~kivy.uix.widget.Widget` properties. You must specify at least a
property name and target value. To use an Animation, follow these steps 

    * Setup an Animation object
    * Use the Animation object on a Widget

Simple animation
----------------

To animate a Widget's x or y position, simply specify the target x/y values
where you want the widget positioned at the end of the animation  

    anim = Animation(x=100, y=100)
    anim.start(widget)

The animation will last for 1 second unless  attr `duration` is specified.
When anim.start() is called, the Widget will move smoothly from the current
x/y position to (100, 100).

Multiple properties and transitions
-----------------------------------

You can animate multiple properties and use built-in or custom transition
functions using  attr `transition` (or the `t=` shortcut). For example,
to animate the position and size using the 'in_quad' transition  

    anim = Animation(x=50, size=(80, 80), t='in_quad')
    anim.start(widget)

Note that the `t=` parameter can be the string name of a method in the
 class `AnimationTransition` class or your own animation function.

Sequential animation
--------------------

To join animations sequentially, use the '+' operator. The following example
will animate to x=50 over 1 second, then animate the size to (80, 80) over the
next two seconds  

    anim = Animation(x=50) + Animation(size=(80, 80), duration=2.)
    anim.start(widget)

Parallel animation
------------------

To join animations in parallel, use the '&' operator. The following example
will animate the position to (80, 10) over 1 second, whilst in parallel
animating the size to (800, 800)  

    anim = Animation(pos=(80, 10))
    anim &= Animation(size=(800, 800), duration=2.)
    anim.start(widget)

Keep in mind that creating overlapping animations on the same property may have
unexpected results. If you want to apply multiple animations to the same
property, you should either schedule them sequentially (via the '+' operator or
using the *on_complete* callback) or cancel previous animations using the
 attr `~Animation.cancel_all` method.

Repeating animation
-------------------

.. versionadded   1.8.0

.. note  
    This is currently only implemented for 'Sequence' animations.

To set an animation to repeat, simply set the  attr `Sequence.repeat`
property to `True`  

    anim = Animation(...) + Animation(...)
    anim.repeat = True
    anim.start(widget)

For flow control of animations such as stopping and cancelling, use the methods
already in place in the animation module.
'
]]--

local __all__ = {'Animation', 'AnimationTransition'}

local sqrt, cos, sin, pi
sin = math.sin
pi = math.pi
cos = math.cos
sqrt = math.sqrt

 
--[[Animation.help.__init__ = 'Create an animation definition that can be used to animate a Widget.

     Parameters 
        `duration` or `d`  float, defaults to 1.
            Duration of the animation, in seconds.
        `transition` or `t`  str or func
            Transition function for animate properties. It can be the name of a
            method from  class `AnimationTransition`.
        `step` or `s`  float
            Step in milliseconds of the animation. Defaults to 0, which means
            the animation is updated for every frame.

            To update the animation less often, set the step value to a float.
            For example, if you want to animate at 30 FPS, use s=1/30.

     Events 
        `on_start`  animation, widget
            Fired when the animation is started on a widget.
        `on_complete`  animation, widget
            Fired when the animation is completed or stopped on a widget.
        `on_progress`  animation, widget, progression
            Fired when the progression of the animation is changing.

    .. versionchanged   1.4.0
       :added s/step parameter.

    .. versionchanged   1.10.0
        The default value of the step parameter was changed from 1/60. to 0.
''
]]--
Animation.__events__ = {'onStart', 'onProgress', 'onComplete'}

function Animation:__init__(kw)
        local kw = kw or {}
        -- Initialize
        self._clock_installed = False
        self._duration = kw.d or kw.duration or 1
        self._transition = kw.t or kw.transition or 'linear'
        self._step = kw.s or kw.step or 0
        kw.step = nil
        kw.s = nil
        kw.t = nil
        kw.transition = nil
        kw.d = nil
        kw.duration = nil
        if type(self._transition) == "string" then
            self._transition = AnimationTransition[self._transition]
            if self._transition == nil then
                error("Invalid transition type. If you don't know what to put just\
                leave it empty.")
            end
        end
        self._animated_properties = kw
        self._widgets = {}
        return self
end

function Animation:dispatch(func,...)
    if type(func) == "string" then
        func = self[func]
    end
    func(...)
end

function Animation:bind(event,func)
    if type(event) == "table" then
        for _,_event in ipairs(event) do
            if lume.isarray(_event) then
                self[_event[1]] = _event[2]
            end
            for name,fun in pairs(_event) do
                self[name] = fun
            end
        end
        for name,fun in pairs(event) do
            self[name] = fun
        end
        return self
    end
    self[event] = func
    return self
end

function Animation:unbind(event)
    if type(event) == "table" then
        for name, fun in ipairs(event) do
            self[name] = null
        end
        return
    end
    self[event] = null
    return self
end
    

--[[function Animation.stop_all(widget, *largs)
        --'''Stop all animations that concern a specific widget / list of
        properties.

        Example  

            anim = Animation(x=50)
            anim.start(widget)

            -- and later
            Animation.stop_all(widget, 'x')
        --'''
        if len(largs) 
            for animation in list(Animation._instances) 
                for x in largs 
                    animation.stop_property(widget, x)
        else 
            for animation in set(Animation._instances) 
                animation.stop(widget)
end 
    
function Animation.cancel_all(widget, *largs)
        --'''Cancel all animations that concern a specific widget / list of
        properties. See  attr `cancel`.

        Example  

            anim = Animation(x=50)
            anim.start(widget)

            -- and later
            Animation.cancel_all(widget, 'x')

        .. versionadded   1.4.0
        --'''
        if len(largs) 
            for animation in list(Animation._instances) 
                for x in largs 
                    animation.cancel_property(widget, x)
        else 
            for animation in set(Animation._instances) 
                animation.cancel(widget)
end
]]--

function Animation:start(widget)
        --'''Start the animation on a widget.
        --'''
        self:stop(widget)
        self:_initialize(widget)
        self:_register()
        self:dispatch('onStart', self,widget)
end

function Animation:stop(widget)
        props = lume.pop(self._widgets,widget.uid)
        
        self.done = True
        if props then
            self:dispatch('onComplete', self,widget)
        end
        self:cancel(widget)
end

function Animation:cancel(widget)
        if not widget.uid then
            return
        end
        self._widgets[widget.uid] = nil
        self:_clock_uninstall()
        if not self._widgets then
            self:_unregister()
        end
end


function Animation:stop_property(widget, prop)
        props = self._widgets.get(widget.uid, None)
        if not props then
            return
        end
        props['properties'][prop] = nil

        -- no more properties to animation ? kill the animation.
        if not props['properties'] then
            self.stop(widget)
        end
end

function Animation:cancel_property(widget, prop)
        props = self._widgets.get(widget.uid, None)
        if not props then 
            return
        end
        lume.pop(props['properties'],prop)

        -- no more properties to animation ? kill the animation.
        if not props['properties'] then
            self.cancel(widget)
        end
end

function Animation:have_properties_to_animate(widget)
        props = self._widgets[widget.uid]
        if props and props['properties'] then
            return True
        end
end
    
--
-- Private
--
function Animation:_register()
        Animation._instances:add(self)
end

function Animation:_unregister()
        if Animation._instances:hasItem(self) then
            Animation._instances:remove(self)
        end
end

function Animation:_initialize(widget)
        if widget.uid == nil then
            widget.uid = getAnID()
        end
        self._widgets[widget.uid] = {
            widget = widget,
            properties = {},
            time =  nil
        }
        local d = self._widgets[widget.uid]
        self.done = False

        -- get current values
        local p = d['properties']
        for key, value in pairs(self._animated_properties) do
        
            local original_value = widget[key]

            if type(original_value) == "table" then
                original_value = lume.copy(original_value)
            end
            p[key] = {original_value, value}
            ::continue::
        end
        -- install clock
        self:_clock_install()
end

function Animation:_clock_install()
        if self._clock_installed then
            return
        end
        self._update_ev = Clock.schedule_interval(self,self._update, self._step)
        self._clock_installed = True
end

function Animation:_clock_uninstall()
        if self._widgets or not self._clock_installed  then
            return
        end
        self._clock_installed = False
        if self._update_ev then
            self._update_ev:cancel()
            self._update_ev = nil
        end
end

function Animation:update(dt)
        if self._clock_installed then
            self:_update(dt)
        end
end

function Animation:_update(dt)
        widgets = self._widgets
        transition = self._transition
        calculate = self._calculate
        for uid,_ in pairs(widgets) do
            anim = widgets[uid]
            widget = anim['widget']

            --[[if isinstance(widget, WeakProxy) and not len(dir(widget)) 
                -- empty proxy, widget is gone. ref  --2458
                self._widgets.pop(uid, None)
                self._clock_uninstall()
                if not self._widgets 
                    self._unregister()
                goto continue
            endp]]--

            if not anim['time'] then
                anim['time'] = 0.
            else
                anim['time'] = anim.time+dt
            end

            -- calculate progression
            if self._duration then
                progress = math.min(1., anim['time'] / self._duration)
            else 
                progress = 1
            end
            t = transition(progress)

            -- apply progression on widget
            for key, values in pairs(anim['properties']) do
                a, b = values[1],values[2]
                value = calculate(self, a, b, t)
                widget[key] = value
            end

            self:dispatch('onProgress', self, widget, progress)

            -- time to stop ?
            if progress >= 1. then
                self:stop(widget)
            end
              ::continue::  
        end
end

function Animation:_calculate(a, b, t) 
        local _calculate = self._calculate
        
        if type(a) == "table" then
            local ip, d = {},{}
            for x,v in ipairs(a) do
                table.insert(ip,_calculate(self,a[x],b[x],t))
            end
            for x,v in pairs(a) do
                d[x] = _calculate(self,a[x],b[x],t)
            end
            if d then
                return d
            else
                return ip
            end

        else 
            return (a * (1. - t)) + (b * t)
        end
end
    --
    -- Default handlers
    --
function Animation:onStart(widget)
end

function Animation:onProgress(widget, progress) 
end

function Animation:onComplete(widget)
end

function  Animation:__add(animation) 
        return Sequence(self, animation)
end

function Animation:__mul(animation) 
        return Parallel(self, animation)
end

Sequence = Animation:extend("Sequence")

function Sequence:__init__(anim1, anim2) 
        Animation.__init__(self)

        --  Repeat the sequence. See 'Repeating animation' in the header
        --  documentation.
        self._repeat = False

        self.anim1 = anim1
        self.anim2 = anim2

        self.anim1:bind({
            {onStart=function(...) self:on_anim1_start(...) end},
            {'onProgress',function(...) self:on_anim1_progress(...) end}
        })
        self.anim2:bind({
            {onComplete=function(...) self:on_anim2_complete(...) end},
            {onProgress=function(...) self:on_anim2_progress(...) end}
        })
end

function Sequence:duration() 
        return self.anim1._duration + self.anim2._duration
end

function Sequence:start(widget)
        self:stop(widget)
        if not widget.uid then
            widget.uid = getAnID()
        end
        self._widgets[widget.uid] = True
        self:_register()
        self.anim1:start(widget)
        self.anim1:bind(onComplete,function(...) self:on_anim1_complete(...) end)
end

function Sequence:stop(widget) 
        self.anim1:stop(widget)
        self.anim2:stop(widget)
        props = lume.pop(self._widgets,widget.uid)
        if props then
            self:dispatch('onComplete', widget)
        end
        self.super.cancel(self,widget)
end

function Sequence:stop_property(widget, prop) 
        self.anim1:stop_property(widget, prop)
        self.anim2:stop_property(widget, prop)
        if (not self.anim1:have_properties_to_animate(widget) and
                not self.anim2:have_properties_to_animate(widget)) then
            self:stop(widget)
        end
end

function Sequence:cancel(widget) 
        self.anim1:cancel(widget)
        self.anim2:cancel(widget)
        self.super.cancel(self,widget)
end

function Sequence:cancel_property(widget, prop)
        self.anim1:cancel_property(widget, prop)
        self.anim2:cancel_property(widget, prop)
        if (not self.anim1:have_properties_to_animate(widget) and
                not self.anim2:have_properties_to_animate(widget)) then
            self:cancel(widget)
        end
end

function Sequence:on_anim1_start(instance, widget) 
        self:dispatch('onStart', widget)
end

function Sequence:on_anim1_complete(instance, widget) 
        self.anim1:unbind(onComplete)
        self.anim2:start(widget)
end

function Sequence:on_anim1_progress(instance, widget, progress)
        self:dispatch('onProgress', widget, progress / 2.)
end

function Sequence:on_anim2_complete(instance, widget) 
        if self._repeat then
            self.anim1:start(widget)
            self.anim1:bind(onComplete,self.on_anim1_complete)
        else 
            self:dispatch('onComplete', widget)
            self:cancel(widget)
        end
end

function Sequence:on_anim2_progress(instance, widget, progress) 
       self:dispatch('onProgress', widget, .5 + progress / 2.)
end

Parallel = Animation:extend("Parallel")

function Parallel:__init__(anim1, anim2) 
        self.super.__init__(self)
        self.anim1 = anim1
        self.anim2 = anim2

        self.anim1:bind(onComplete,function(...) self:on_anim_complete(...) end)
        self.anim2:bind(onComplete,function(...) self:on_anim_complete(...) end)
end
    
function Parallel:duration() 
        return lume.max(self.anim1.duration, self.anim2.duration)
end

function Parallel:start(widget) 
        self:stop(widget)
        self.anim1:start(widget)
        self.anim2:start(widget)
        self._widgets[widget.uid] = {complete=0}
        self:_register()
        self:dispatch('onStart', widget)
end

function Parallel:stop(widget) 
        self.anim1:stop(widget)
        self.anim2:stop(widget)
        props = lume.pop(self._widgets,widget.uid)
        if props then
            self:dispatch('onComplete', widget)
        end
        self.super.cancel(self,widget)

end

function Parallel:stop_property(widget, prop) 
        self.anim1:stop_property(widget, prop)
        self.anim2:stop_property(widget, prop)
        if (not self.anim1:have_properties_to_animate(widget) and
                not self.anim2:have_properties_to_animate(widget)) then
            self:stop(widget)
        end
end

function Parallel:cancel(widget) 
        self.anim1:cancel(widget)
        self.anim2:cancel(widget)
        self.super.cancel(self,widget)

end

function Parallel:on_anim_complete(instance, widget) 
        self._widgets[widget.uid]['complete'] = self._widgets[widget.uid].complete + 1
        if self._widgets[widget.uid]['complete'] == 2  then
            self:stop(widget)
        end
end


AnimationTransition = { 
    doc= 'Collection of animation functions to be used with the Animation object.\
    Easing Functions ported to Lua from Kivy from the Clutter Project\
    https //developer.gnome.org/clutter/stable/ClutterAlpha.html\
\
    The `progress` parameter in each animation function is in the range 0-1.\
    ',

   
    linear = function(progress)
        --'''.. image   images/anim_linear.png--'''
        return progress
    end,

    
    in_quad = function(progress)
        --'''.. image   images/anim_in_quad.png
        --'''
        return progress * progress
    end,

    out_quad = function(progress)
        --'''.. image   images/anim_out_quad.png
        --'''
        return -1.0 * progress *  (progress - 2.0)
    end,

    in_out_quad = function(progress)
        --'''.. image   images/anim_in_out_quad.png
        --'''
        p = progress * 2
        if p < 1 then
            return 0.5 * p * p
        end
        p = p - 1.0
        return -0.5 * (p * (p - 2.0) - 1.0)
    end,
    
    in_cubic = function(progress)
        --'''.. image   images/anim_in_cubic.png
        --'''
        return progress * progress * progress
    end,


    out_cubic = function(progress)
        --'''.. image   images/anim_out_cubic.png
        --'''
        p = progress - 1.0
        return p * p * p + 1.0
    end,


    in_out_cubic = function(progress) 
        --'''.. image   images/anim_in_out_cubic.png
        --'''
        p = progress * 2
        if p < 1 then 
            return 0.5 * p * p * p
        end
        p = p - 2
        return 0.5 * (p * p * p + 2.0)
    end,

    in_quart = function(progress)
        --'''.. image   images/anim_in_quart.png
        --'''
        return progress * progress * progress * progress
    end,

    out_quart = function(progress)
        --'''.. image   images/anim_out_quart.png
        --'''
        p = progress - 1.0
        return -1.0 * (p * p * p * p - 1.0)
    end,

    in_out_quart = function(progress)
        --'''.. image   images/anim_in_out_quart.png
        --'''
        p = progress * 2
        if p < 1 then 
            return 0.5 * p * p * p * p
        end
        p = p - 2
        return -0.5 * (p * p * p * p - 2.0)
    end,

    in_quint = function(progress) 
        --'''.. image   images/anim_in_quint.png
        --'''
        return progress * progress * progress * progress * progress
    end,

    ot_quint = function(progress) 
        --'''.. image   images/anim_out_quint.png
        --'''
        p = progress - 1.0
        return p * p * p * p * p + 1.0
    end,

    in_out_quint = function(progress) 
        --'''.. image   images/anim_in_out_quint.png
        --'''
        p = progress * 2
        if p < 1 then 
            return 0.5 * p * p * p * p * p
        end
        p = p - 2.0
        return 0.5 * (p * p * p * p * p + 2.0)
    end,

    in_sine = function(progress) 
        --'''.. image   images/anim_in_sine.png
        --'''
        return -1.0 * cos (progress * (pi / 2.0)) + 1.0
    end,

    ot_sine = function(progress) 
        --'''.. image   images/anim_out_sine.png
        --'''
        return sin (progress * (pi / 2.0))
    end,

    in_out_sine = function(progress) 
        --'''.. image   images/anim_in_out_sine.png
        --'''
        return -0.5 * (cos(pi * progress) - 1.0)
    end,

    in_expo = function(progress) 
        --'''.. image   images/anim_in_expo.png
        --'''
        if progress == 0 then
            return 0.0
        end
        return pow(2, 10 *  (progress - 1.0))
    end,

    out_expo = function(progress) 
        --'''.. image   images/anim_out_expo.png
        --'''
        if progress == 1.0 then
            return 1.0
        end
        return -pow(2, -10 * progress) + 1.0
    end,

    in_out_expo = function(progress) 
        --'''.. image   images/anim_in_out_expo.png
        --'''
        if progress == 0 then
            return 0.0
        end
        if progress == 1. then 
            return 1.0
        end
        p = progress * 2
        if p < 1 then 
            return 0.5 * pow(2, 10 * (p - 1.0))
        end
        p = p - 1.0
        return 0.5 * (-pow(2, -10 * p) + 2.0)
    end,

    in_circ = function(progress) 
        --'''.. image   images/anim_in_circ.png
        --'''
        return -1.0 * (sqrt(1.0 - progress * progress) - 1.0)
    end,

    ot_circ = function(progress) 
        --'''.. image   images/anim_out_circ.png
        --'''
        p = progress - 1.0
        return sqrt(1.0 - p * p)
    end,

    in_out_circ = function(progress) 
        --'''.. image   images/anim_in_out_circ.png
        --'''
        p = progress * 2
        if p < 1 then 
            return -0.5 * (sqrt(1.0 - p * p) - 1.0)
        end
        p = p - 2.0
        return 0.5 * (sqrt(1.0 - p * p) + 1.0)
    end,

    in_elastic = function(progress) 
        --'''.. image   images/anim_in_elastic.png
        --'''
        p = .3
        s = p / 4.0
        q = progress
        if q == 1 then
            return 1.0
        end
        q = q - 1.0
        return -(pow(2, 10 * q) * sin((q - s) * (2 * pi) / p))
    end,

    ot_elastic = function(progress) 
        --'''.. image   images/anim_out_elastic.png
        --'''
        p = .3
        s = p / 4.0
        q = progress
        if q == 1 then
            return 1.0
        end
        return pow(2, -10 * q) * sin((q - s) * (2 * pi) / p) + 1.0
    end,

    in_out_elastic = function(progress) 
        --'''.. image   images/anim_in_out_elastic.png
        --'''
        p = .3 * 1.5
        s = p / 4.0
        q = progress * 2
        if q == 2 then
            return 1.0
        end
        if q < 1 then
            q = q - 1.0
            return -.5 * (pow(2, 10 * q) * sin((q - s) * (2.0 * pi) / p))
        else 
            q = q - 1.0
            return pow(2, -10 * q) * sin((q - s) * (2.0 * pi) / p) * .5 + 1.0
        end
    end,

    in_back = function(progress) 
        --'''.. image   images/anim_in_back.png
        --'''
        return progress * progress * ((1.70158 + 1.0) * progress - 1.70158)
    end,

    ot_back = function(progress) 
        --'''.. image   images/anim_out_back.png
        --'''
        p = progress - 1.0
        return p * p * ((1.70158 + 1) * p + 1.70158) + 1.0
    end,

    in_out_back = function(progress) 
        --'''.. image   images/anim_in_out_back.png
        --'''
        p = progress * 2.
        s = 1.70158 * 1.525
        if p < 1 then 
            return 0.5 * (p * p * ((s + 1.0) * p - s))
        end
        p = p - 2.0
        return 0.5 * (p * p * ((s + 1.0) * p + s) + 2.0)
    end,

    _out_bounce_internal = function(t, d) 
        p = t / d
        if p < (1.0 / 2.75) then
            return 7.5625 * p * p
        elseif p < (2.0 / 2.75) then
            p = p - (1.5 / 2.75)
            return 7.5625 * p * p + .75
        elseif p < (2.5 / 2.75) then
            p = p - (2.25 / 2.75)
            return 7.5625 * p * p + .9375
        else 
            p = p - (2.625 / 2.75)
            return 7.5625 * p * p + .984375
        end
    end,

    _in_bounce_internal = function(t, d) 
        return 1.0 - AnimationTransition._out_bounce_internal(d - t, d)
    end,

    in_bounce = function(progress) 
        --'''.. image   images/anim_in_bounce.png
        --'''
        return AnimationTransition._in_bounce_internal(progress, 1.)
    end,

    ot_bounce = function(progress) 
        --'''.. image   images/anim_out_bounce.png
        --'''
        return AnimationTransition._out_bounce_internal(progress, 1.)
    end,

    in_out_bounce = function(progress) 
        --'''.. image   images/anim_in_out_bounce.png
        --'''
        p = progress * 2.
        if p < 1. then
            return AnimationTransition._in_bounce_internal(p, 1.) * .5
        end
        return AnimationTransition._out_bounce_internal(p - 1., 1.) * .5 + .5
    end
}

Animation.AnimationTransitions = AnimationTransition
return Animation